"""Module grouping tests for the search grondmonster module."""

from owslib.fes2 import PropertyIsEqualTo

from pydov.search.grondmonster import GrondmonsterSearch
from pydov.types.fields import ReturnFieldList
from pydov.types.grondmonster import Grondmonster
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/grondmonster/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondmonster/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondmonster/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/grondmonster/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondmonster/feature.xml'
location_dov_xml = 'tests/data/types/grondmonster/grondmonster.xml'
location_xsd_base = 'tests/data/types/grondmonster/xsd_*.xml'


class TestGrondmonsterSearch(AbstractTestSearch):

    search_instance = GrondmonsterSearch()
    datatype_class = Grondmonster

    valid_query_single = PropertyIsEqualTo(propertyname='boornummer',
                                           literal='GEO-04/024-B6')

    inexistent_field = 'onbestaand'
    wfs_field = 'boornummer'
    xml_field = 'astm_naam'

    valid_returnfields = ReturnFieldList.from_field_names('pkey_grondmonster', 'boornummer')
    valid_returnfields_subtype = ReturnFieldList.from_field_names(
        'pkey_grondmonster', 'boornummer', 'diameter')
    valid_returnfields_extra = ReturnFieldList.from_field_names('pkey_grondmonster', 'korrelverdeling')

    df_default_columns = [
        'pkey_grondmonster', 'naam', 'pkey_boring', 'boornummer',
        'datum', 'x', 'y', 'gemeente', 'diepte_van_m', 'diepte_tot_m',
        'peil_van_mtaw', 'peil_tot_mtaw', 'monstertype', 'astm_naam',
        'grondsoort_bggg', 'humusgehalte', 'kalkgehalte',
        'uitrolgrens', 'vloeigrens', 'glauconiet_totaal',
        'korrelvolumemassa', 'volumemassa', 'watergehalte',
        'diameter', 'fractie', 'methode']

    def test_search_xmlresolving(self, mp_get_schema,
                                 mp_remote_describefeaturetype,
                                 mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_grondmonster', 'boornummer', 'humusgehalte',
                           'methode'))

        assert df.humusgehalte[0] == 4.7
        assert df.methode[0] == 'ZEEFPROEF'

    def test_issue_285(self, mp_get_schema,
                       mp_remote_describefeaturetype,
                       mp_remote_wfs_feature, mp_dov_xml):
        """Test whether korrelvolumemassa, volumemassa and watergehalte
        resolve correctly.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.
        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=(
                'boornummer',
                'naam',
                'korrelvolumemassa',
                'volumemassa',
                'watergehalte'))

        assert round(df.korrelvolumemassa[0], 1) == 2.6
        assert round(df.volumemassa[0], 1) == 1.6
        assert round(df.watergehalte[0], 1) == 57.7



"""Module grouping tests for the pydov.util.owsutil module, without all
module scoped monkeypatches."""
import copy

from pydov.util import owsutil


class TestOwsutilNoMP(object):
    """Class grouping tests for the pydov.util.owsutil module, without all
    module scoped monkeypatches.."""

    def test_get_remote_metadata_nometadataurls(self, wfs):
        """Test the owsutil.get_remote_metadata method when the WFS layer
        missed metadata URLs.

        Test whether None is returned.

        Parameters
        ----------
        wfs : pytest.fixture returning owslib.wfs.WebFeatureService
            WebFeatureService based on the local GetCapabilities.

        """
        contents = copy.deepcopy(wfs.contents)
        contentmetadata = contents['dov-pub:Boringen']
        contentmetadata.metadataUrls = []
        assert owsutil.get_remote_metadata(contentmetadata) is None

    def test_get_remote_metadata_nometadata(self, wfs, mp_geonetwork_broken):
        """Test the owsutil.get_remote_metadata method when the remote metadata
        could not be found.

        Test whether None is returned.

        Parameters
        ----------
        wfs : pytest.fixture returning owslib.wfs.WebFeatureService
            WebFeatureService based on the local GetCapabilities.

        """
        contents = copy.deepcopy(wfs.contents)
        contentmetadata = contents['dov-pub:Boringen']
        assert owsutil.get_remote_metadata(contentmetadata) is None



"""Module grouping tests for the pydov.types.bodemlocatie module."""

from pydov.types.bodemlocatie import Bodemlocatie
from pydov.types.fields import ReturnFieldList
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/bodemlocatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemlocatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemlocatie/bodemlocatie.xml'


class TestBodemlocatie(AbstractTestTypes):
    """Class grouping tests for the pydov.types.bodemlocatie.Bodemlocatie class."""

    datatype_class = Bodemlocatie
    namespace = 'https://www.dov.vlaanderen.be/bodem'
    pkey_base = build_dov_url('data/bodemlocatie/')

    field_names = [
        'pkey_bodemlocatie', 'pkey_bodemsite',
        'naam', 'type', 'waarnemingsdatum', 'doel',
        'x', 'y', 'mv_mtaw',
        'erfgoed', 'bodemstreek',
        'invoerdatum', 'educatieve_waarde'
    ]
    field_names_subtypes = []
    field_names_nosubtypes = [
        'pkey_bodemlocatie', 'pkey_bodemsite',
        'naam', 'type', 'waarnemingsdatum', 'doel',
        'x', 'y', 'mv_mtaw',
        'erfgoed', 'bodemstreek',
        'invoerdatum', 'educatieve_waarde'
    ]

    valid_returnfields = ReturnFieldList.from_field_names('pkey_bodemlocatie', 'naam')
    valid_returnfields_subtype = ReturnFieldList.from_field_names('pkey_bodemlocatie', 'naam')

    inexistent_field = 'onbestaand'

    def test_get_field_names_wrongreturnfields_nosubtypes(self):
        """Test the get_field_names method when disabling subtypes
        and including an otherwise valid return field.

        Test whether an InvalidFieldError is raised.

        Override because bodemlocatie has no subtypes.
        """
        assert True



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV borehole data."""
from pydov.search.abstract import AbstractSearch
from pydov.types.grondmonster import Grondmonster


class GrondmonsterSearch(AbstractSearch):
    """Search class to retrieve the grain size distribution of
    ground samples ('grondmonster')"""

    def __init__(self, objecttype=Grondmonster):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Grondmonster type.
            Optional: defaults to the Grondmonster type containing the fields
            described in the documentation.

        """
        super(GrondmonsterSearch, self).\
            __init__('boringen:grondmonsters', objecttype)



# -*- coding: utf-8 -*-
"""Module containing tests for generic fields regarding location (ligging)."""


import numpy as np
from owslib.etree import etree

from pydov.types.boring import Boring
from pydov.types.ligging import MvMtawField


class TestMvMtawField:
    def test_original_xml(self):
        """Test whether the correct mv_mtaw value is returned for the original XML schema."""
        mv_mtaw_field = MvMtawField('Maaiveldhoogte in mTAW op dag dat de boring uitgevoerd werd.')

        boring_tree = etree.fromstring(
            """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <ns4:dov-schema xmlns:gml="http://www.opengis.net/gml/3.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:ns4="http://kern.schemas.dov.vlaanderen.be" xsi:schemaLocation="http://kern.schemas.dov.vlaanderen.be https://www.dov.vlaanderen.be/xdov/schema/latest/xsd/kern/dov.xsd">
                    <boring>
                        <identificatie>kb29d84e-B574</identificatie>
                        <dataidentifier>
                            <permkey>2016-122561</permkey>
                            <uri>https://oefen.dov.vlaanderen.be/data/boring/2016-122561</uri>
                        </dataidentifier>
                        <xy>
                            <x>92424.00</x>
                            <y>170752.00</y>
                            <betrouwbaarheid>onbekend</betrouwbaarheid>
                            <methode_opmeten>gedigitaliseerd op topokaart</methode_opmeten>
                            <origine_opmeten>
                                <naam>Universiteit Gent</naam>
                            </origine_opmeten>
                        </xy>
                        <gemeente>45061</gemeente>
                        <oorspronkelijk_maaiveld>
                            <waarde>40.00</waarde>
                            <betrouwbaarheid>onbekend</betrouwbaarheid>
                            <methode_opmeten>afgeleid van topokaart</methode_opmeten>
                            <origine_opmeten>
                                <naam>Universiteit Gent</naam>
                            </origine_opmeten>
                        </oorspronkelijk_maaiveld>
                        <start_tov_maaiveld>
                            <gestart_op>MAAIVELD</gestart_op>
                        </start_tov_maaiveld>
                        <diepte_van>0.00</diepte_van>
                        <diepte_tot>0.00</diepte_tot>
                        <wet_kader>
                            <niet_ingedeeld/>
                        </wet_kader>
                        <uitvoerder>
                            <naam>onbekend</naam>
                        </uitvoerder>
                        <opdrachtgever>
                            <naam>onbekend</naam>
                        </opdrachtgever>
                        <dataleverancier>
                            <naam>onbekend</naam>
                        </dataleverancier>
                        <boorgatmeting>
                            <uitgevoerd>false</uitgevoerd>
                        </boorgatmeting>
                        <stalen>
                            <bewaard>false</bewaard>
                        </stalen>
                        <details>
                            <boormethode>
                                <van>0.00</van>
                                <tot>0.00</tot>
                                <methode>onbekend</methode>
                            </boormethode>
                        </details>
                    </boring>
                </ns4:dov-schema>""".encode('utf8'))

        mv_mtaw = mv_mtaw_field.calculate(Boring, boring_tree)
        assert mv_mtaw == 40.0

    def test_new_xml_maaiveld(self):
        """Test whether the correct mv_mtaw value is returned for the new XML schema."""
        mv_mtaw_field = MvMtawField('Maaiveldhoogte in mTAW op dag dat de boring uitgevoerd werd.')

        boring_tree = etree.fromstring(
            """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <ns4:dov-schema xmlns:gml="http://www.opengis.net/gml/3.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:ns4="http://kern.schemas.dov.vlaanderen.be" xsi:schemaLocation="http://kern.schemas.dov.vlaanderen.be https://www.dov.vlaanderen.be/xdov/schema/latest/xsd/kern/dov.xsd">
                    <boring>
                        <identificatie>kb29d84e-B574</identificatie>
                        <dataidentifier>
                            <permkey>2016-122561</permkey>
                            <uri>https://ontwikkel.dov.vlaanderen.be/data/boring/2016-122561</uri>
                        </dataidentifier>
                        <ligging>
                            <gml:Point srsName="urn:ogc:def:crs:EPSG::6190" srsDimension="3">
                                <gml:pos>92424.0 170752.0 40.0</gml:pos>
                            </gml:Point>
                            <metadata_locatiebepaling>
                                <methode>gedigitaliseerd op topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                            </metadata_locatiebepaling>
                            <metadata_hoogtebepaling>
                                <methode>afgeleid van topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                                <referentiepunt_type>Maaiveld</referentiepunt_type>
                            </metadata_hoogtebepaling>
                        </ligging>
                        <diepte_van>0.00</diepte_van>
                        <diepte_tot>0.00</diepte_tot>
                        <wet_kader>
                            <niet_ingedeeld/>
                        </wet_kader>
                        <uitvoerder>
                            <naam>onbekend</naam>
                        </uitvoerder>
                        <opdrachtgever>
                            <naam>onbekend</naam>
                        </opdrachtgever>
                        <dataleverancier>
                            <naam>onbekend</naam>
                        </dataleverancier>
                        <boorgatmeting>
                            <uitgevoerd>false</uitgevoerd>
                        </boorgatmeting>
                        <stalen>
                            <bewaard>true</bewaard>
                        </stalen>
                        <details>
                            <boormethode>
                                <van>0.00</van>
                                <tot>0.00</tot>
                                <methode>onbekend</methode>
                            </boormethode>
                        </details>
                        <opdracht>CTE verzamel Opdrachtgever</opdracht>
                    </boring>
                </ns4:dov-schema>""".encode('utf8'))

        mv_mtaw = mv_mtaw_field.calculate(Boring, boring_tree)
        assert mv_mtaw == 40.0

    def test_new_xml_nomaaiveld(self):
        """Test whether NaN is returned when the reference point is not Maaiveld."""
        mv_mtaw_field = MvMtawField('Maaiveldhoogte in mTAW op dag dat de boring uitgevoerd werd.')

        boring_tree = etree.fromstring(
            """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <ns4:dov-schema xmlns:gml="http://www.opengis.net/gml/3.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:ns4="http://kern.schemas.dov.vlaanderen.be" xsi:schemaLocation="http://kern.schemas.dov.vlaanderen.be https://www.dov.vlaanderen.be/xdov/schema/latest/xsd/kern/dov.xsd">
                    <boring>
                        <identificatie>kb29d84e-B574</identificatie>
                        <dataidentifier>
                            <permkey>2016-122561</permkey>
                            <uri>https://ontwikkel.dov.vlaanderen.be/data/boring/2016-122561</uri>
                        </dataidentifier>
                        <ligging>
                            <gml:Point srsName="urn:ogc:def:crs:EPSG::6190" srsDimension="3">
                                <gml:pos>92424.0 170752.0 40.0</gml:pos>
                            </gml:Point>
                            <metadata_locatiebepaling>
                                <methode>gedigitaliseerd op topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                            </metadata_locatiebepaling>
                            <metadata_hoogtebepaling>
                                <methode>afgeleid van topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                                <referentiepunt_type>Aanvangspeil</referentiepunt_type>
                            </metadata_hoogtebepaling>
                        </ligging>
                        <diepte_van>0.00</diepte_van>
                        <diepte_tot>0.00</diepte_tot>
                        <wet_kader>
                            <niet_ingedeeld/>
                        </wet_kader>
                        <uitvoerder>
                            <naam>onbekend</naam>
                        </uitvoerder>
                        <opdrachtgever>
                            <naam>onbekend</naam>
                        </opdrachtgever>
                        <dataleverancier>
                            <naam>onbekend</naam>
                        </dataleverancier>
                        <boorgatmeting>
                            <uitgevoerd>false</uitgevoerd>
                        </boorgatmeting>
                        <stalen>
                            <bewaard>true</bewaard>
                        </stalen>
                        <details>
                            <boormethode>
                                <van>0.00</van>
                                <tot>0.00</tot>
                                <methode>onbekend</methode>
                            </boormethode>
                        </details>
                        <opdracht>CTE verzamel Opdrachtgever</opdracht>
                    </boring>
                </ns4:dov-schema>""".encode('utf8'))

        mv_mtaw = mv_mtaw_field.calculate(Boring, boring_tree)
        assert mv_mtaw is np.nan

    def test_new_xml_noheight(self):
        """Test whether NaN is returned when the point does not have height information."""
        mv_mtaw_field = MvMtawField('Maaiveldhoogte in mTAW op dag dat de boring uitgevoerd werd.')

        boring_tree = etree.fromstring(
            """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
                <ns4:dov-schema xmlns:gml="http://www.opengis.net/gml/3.2" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:ns4="http://kern.schemas.dov.vlaanderen.be" xsi:schemaLocation="http://kern.schemas.dov.vlaanderen.be https://www.dov.vlaanderen.be/xdov/schema/latest/xsd/kern/dov.xsd">
                    <boring>
                        <identificatie>kb29d84e-B574</identificatie>
                        <dataidentifier>
                            <permkey>2016-122561</permkey>
                            <uri>https://ontwikkel.dov.vlaanderen.be/data/boring/2016-122561</uri>
                        </dataidentifier>
                        <ligging>
                            <gml:Point srsName="urn:ogc:def:crs:EPSG::31370" srsDimension="2">
                                <gml:pos>92424.0 170752.0</gml:pos>
                            </gml:Point>
                            <metadata_locatiebepaling>
                                <methode>gedigitaliseerd op topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                            </metadata_locatiebepaling>
                            <metadata_hoogtebepaling>
                                <methode>afgeleid van topokaart</methode>
                                <uitvoerder>
                                    <naam>Universiteit Gent</naam>
                                </uitvoerder>
                                <betrouwbaarheid>onbekend</betrouwbaarheid>
                                <referentiepunt_type>Maaiveld</referentiepunt_type>
                            </metadata_hoogtebepaling>
                        </ligging>
                        <diepte_van>0.00</diepte_van>
                        <diepte_tot>0.00</diepte_tot>
                        <wet_kader>
                            <niet_ingedeeld/>
                        </wet_kader>
                        <uitvoerder>
                            <naam>onbekend</naam>
                        </uitvoerder>
                        <opdrachtgever>
                            <naam>onbekend</naam>
                        </opdrachtgever>
                        <dataleverancier>
                            <naam>onbekend</naam>
                        </dataleverancier>
                        <boorgatmeting>
                            <uitgevoerd>false</uitgevoerd>
                        </boorgatmeting>
                        <stalen>
                            <bewaard>true</bewaard>
                        </stalen>
                        <details>
                            <boormethode>
                                <van>0.00</van>
                                <tot>0.00</tot>
                                <methode>onbekend</methode>
                            </boormethode>
                        </details>
                        <opdracht>CTE verzamel Opdrachtgever</opdracht>
                    </boring>
                </ns4:dov-schema>""".encode('utf8'))

        mv_mtaw = mv_mtaw_field.calculate(Boring, boring_tree)
        assert mv_mtaw is np.nan



import pydov
import pytest
import os
import re

from owslib.etree import etree
from owslib.fes2 import PropertyIsGreaterThanOrEqualTo
from pydov.search.boring import BoringSearch
from pydov.util import owsutil

location_md_metadata = 'tests/data/types/boring/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/boring/fc_featurecatalogue.xml'
location_wfs_capabilities = 'tests/data/wfs_paging/wfscapabilities.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/boring/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/wfs_paging/wfsgetfeature.xml'

page_size = 10


@pytest.fixture
def wfs_capabilities():
    """PyTest fixture providing the WFS GetCapabilities response based on a 
    local copy. Adjusts the CountDefault parameter to the configured 
    `page_size` in this module.

    Returns
    -------
    bytes
        WFS 2.0.0 GetCapabilities response.
    """
    with open('tests/data/util/owsutil/wfscapabilities.xml', 'r',
              encoding='utf-8') as f:

        data = re.sub(r'<ows:Constraint name="CountDefault"><ows:NoValues/>'
                      r'<ows:DefaultValue>[1-9][0-9]*</ows:DefaultValue>'
                      r'</ows:Constraint>',
                      '<ows:Constraint name="CountDefault"><ows:NoValues/>'
                      f'<ows:DefaultValue>{page_size}</ows:DefaultValue>'
                      '</ows:Constraint>',
                      f.read())

        if not isinstance(data, bytes):
            data = data.encode('utf-8')

    return data


@pytest.fixture
def mp_wfs_max_features(monkeypatch, wfs_capabilities):
    """Monkeypatch the call to get the WFS capabilities maximum features.

    This ensures the maximum features are read from the testdata capabilities.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    wfs_capabilities : pytest.fixture
        PyTest fixture providing the WFS capabilities from the testdata.

    """
    max_features = owsutil.get_wfs_max_features(wfs_capabilities)

    def __get_wfs_max_features(*args, **kwargs):
        return max_features

    monkeypatch.setattr(pydov.util.owsutil,
                        'get_wfs_max_features',
                        __get_wfs_max_features)


@pytest.fixture
def mp_remote_wfs_paged_feature(monkeypatch, request):
    """Monkeypatch the call to get WFS features.

    This monkeypatch requires a module variable ``location_wfs_getfeature``
    with the path to the wfs_getfeature file on disk.

    This patch uses both the start_index as well as the max_features from the
    WFS request to ensure the right page and number of results are returned.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_remote_wfs_feature(*args, **kwargs):
        start_index = int(kwargs['get_feature_request'].get('startIndex'))

        count = kwargs['get_feature_request'].get('count')
        count = int(count) if count is not None else None

        file_path = getattr(request.module, "location_wfs_getfeature")
        file_path = (file_path[:file_path.rfind(
            '.')] + f'_{start_index//page_size}' +
            file_path[file_path.rfind('.'):])

        if not os.path.exists(file_path):
            raise RuntimeError(
                f'Error reading testdata: {file_path} does not exist.')

        with open(file_path, 'r') as f:
            data = f.read()

            if count is not None:
                tree = etree.fromstring(data)
                members = tree.findall(
                    './/{http://www.opengis.net/wfs/2.0}member')
                for i in range(len(members)):
                    if i >= count:
                        tree.remove(members[i])
                data = etree.tostring(tree)

            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.owsutil,
                        'wfs_get_feature',
                        __get_remote_wfs_feature)


class TestSearchWfsPaging(object):
    """Class grouping tests regarding WFS paging."""

    def test_all_features_paged(
            self, mp_wfs, mp_get_schema, mp_remote_describefeaturetype,
            mp_wfs_max_features, mp_remote_wfs_paged_feature):
        """Test a search needing multiple pages.

        Test whether the number of returned features is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_wfs_max_features : pytest.fixture
            Monkeypatch the call to get the maximum features from the
            capabilities.
        mp_remote_wfs_paged_feature : pytest.fixture
            Monkeypatch the call to the remote WFS GetFeature, with support
            for paging.

        """
        s = BoringSearch()
        df = s.search(query=PropertyIsGreaterThanOrEqualTo(
            'diepte_tot_m', '0'), return_fields=['pkey_boring'])
        assert len(df) == 20

    def test_max_features_first_page(
            self, mp_wfs, mp_get_schema, mp_remote_describefeaturetype,
            mp_wfs_max_features, mp_remote_wfs_paged_feature):
        """Test a search using max_features where the maximum is less than the
        WFS page size.

        Test whether the number of returned features is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_wfs_max_features : pytest.fixture
            Monkeypatch the call to get the maximum features from the
            capabilities.
        mp_remote_wfs_paged_feature : pytest.fixture
            Monkeypatch the call to the remote WFS GetFeature, with support
            for paging.

        """
        s = BoringSearch()
        df = s.search(return_fields=['pkey_boring'], max_features=5)
        assert len(df) == 5

    def test_max_features_paged(
            self, mp_wfs, mp_get_schema, mp_remote_describefeaturetype,
            mp_wfs_max_features, mp_remote_wfs_paged_feature):
        """Test a search using max_features where the maximum is greater than
        the WFS page size.

        Test whether the number of returned features is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_wfs_max_features : pytest.fixture
            Monkeypatch the call to get the maximum features from the
            capabilities.
        mp_remote_wfs_paged_feature : pytest.fixture
            Monkeypatch the call to the remote WFS GetFeature, with support
            for paging.

        """
        s = BoringSearch()
        df = s.search(return_fields=['pkey_boring'], max_features=15)
        assert len(df) == 15



"""Module grouping session scoped PyTest fixtures."""

import datetime
import glob
import os
import tempfile

import pytest
from owslib.etree import etree
from owslib.feature.schema import _construct_schema, _get_elements
from owslib.feature.wfs200 import ContentMetadata
from owslib.iso import MD_Metadata
from owslib.util import ResponseWrapper, findall
from owslib.wfs import WebFeatureService

import pydov
from pydov import Hooks
from pydov.util import owsutil
from pydov.util.caching import GzipTextFileCache, PlainTextFileCache
from pydov.util.dovutil import build_dov_url


def pytest_runtest_setup():
    pydov.hooks = Hooks()


def pytest_configure(config):
    config.addinivalue_line("markers",
                            "online: mark test that requires internet access")


@pytest.fixture
def wfs_capabilities():
    """PyTest fixture providing the WFS GetCapabilities response based on a 
    local copy.

    Returns
    -------
    bytes
        WFS 2.0.0 GetCapabilities response.
    """
    with open('tests/data/util/owsutil/wfscapabilities.xml', 'r',
              encoding='utf-8') as f:
        data = f.read()
        if not isinstance(data, bytes):
            data = data.encode('utf-8')
    return data


@pytest.fixture
def mp_wfs(monkeypatch, wfs_capabilities):
    """Monkeypatch the call to the remote GetCapabilities request.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    wfs_capabilities: pytest.fixture
        Fixture providing WFS capabilities response.

    """
    def read(*args, **kwargs):
        return wfs_capabilities

    monkeypatch.setattr(
        owsutil, 'get_wfs_capabilities', read)


@pytest.fixture
def wfs(wfs_capabilities):
    """PyTest fixture providing an instance of a WebFeatureService based on
    a local copy of a GetCapabilities request.

    Parameters
    ----------
    wfs_capabilities: pytest.fixture
        Fixture providing WFS capabilities response.

    Returns
    -------
    owslib.wfs.WebFeatureService
        WebFeatureService based on the local GetCapabilities.

    """
    return WebFeatureService(
        url=build_dov_url('geoserver/wfs'), version="2.0.0",
        xml=wfs_capabilities)


@pytest.fixture
def mp_remote_fc_notfound(monkeypatch):
    """Monkeypatch the call to get an inexistent remote featurecatalogue.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.

    """
    def __get_remote_fc(*args, **kwargs):
        with open('tests/data/util/owsutil/fc_featurecatalogue_notfound.xml',
                  'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.owsutil, '__get_remote_fc', __get_remote_fc)


@pytest.fixture
def mp_remote_md(wfs, monkeypatch, request):
    """Monkeypatch the call to get the remote metadata of the layer.

    This monkeypatch requires a module variable ``location_md_metadata``
    with the path to the md_metadata file on disk.

    Parameters
    ----------
    wfs : pytest.fixture returning owslib.wfs.WebFeatureService
        WebFeatureService based on the local GetCapabilities.
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_remote_md(*args, **kwargs):
        file_path = getattr(request.module, "location_md_metadata")
        with open(file_path, 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return MD_Metadata(etree.fromstring(data).find(
            './{http://www.isotc211.org/2005/gmd}MD_Metadata'))

    monkeypatch.setattr(pydov.util.owsutil, 'get_remote_metadata',
                        __get_remote_md)


@pytest.fixture
def mp_remote_fc(monkeypatch, request):
    """Monkeypatch the call to get the remote feature catalogue.

    This monkeypatch requires a module variable
    ``location_fc_featurecatalogue`` with the path to the fc_featurecatalogue
    file on disk.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_remote_fc(*args, **kwargs):
        file_path = getattr(request.module, "location_fc_featurecatalogue")
        with open(file_path, 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.owsutil, '__get_remote_fc',
                        __get_remote_fc)


@pytest.fixture
def mp_remote_describefeaturetype(monkeypatch, request):
    """Monkeypatch the call to a remote DescribeFeatureType.

    This monkeypatch requires a module variable
    ``location_wfs_describefeaturetype`` with the path to the
    wfs_describefeaturetype file on disk.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_remote_describefeaturetype(*args, **kwargs):
        file_path = getattr(request.module, "location_wfs_describefeaturetype")
        with open(file_path, 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.owsutil,
                        '__get_remote_describefeaturetype',
                        __get_remote_describefeaturetype)


@pytest.fixture
def mp_get_schema(monkeypatch, request):
    """Monkeypatch the call to the get_schema method.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_schema(*args, **kwargs):
        file_path = getattr(request.module, "location_wfs_describefeaturetype")
        with open(file_path, 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')

        root = etree.fromstring(data)

        typename = root.find(
            './{http://www.w3.org/2001/XMLSchema}element').get('name')

        if ":" in typename:
            typename = typename.split(":")[1]
        type_element = findall(
            root,
            "{http://www.w3.org/2001/XMLSchema}element",
            attribute_name="name",
            attribute_value=typename,
        )[0]
        complex_type = type_element.attrib["type"].split(":")[1]
        elements = _get_elements(complex_type, root)
        nsmap = None
        if hasattr(root, "nsmap"):
            nsmap = root.nsmap
        return _construct_schema(elements, nsmap)

    monkeypatch.setattr(pydov.search.abstract.AbstractSearch, '_get_schema',
                        __get_schema)


@pytest.fixture
def wfs_getfeature(request):
    """PyTest fixture providing a WFS GetFeature response.

    This monkeypatch requires a module variable ``location_wfs_getfeature``
    with the path to the wfs_getfeature file on disk.

    Parameters
    ----------
    request : pytest.fixture
        PyTest fixture providing request context.

    Returns
    -------
    str
        WFS response of a GetFeature call to the dov-pub:Boringen layer.

    """
    file_path = getattr(request.module, "location_wfs_getfeature")
    with open(file_path, 'r') as f:
        data = f.read()
        return data


@pytest.fixture
def wfs_feature(request):
    """PyTest fixture providing an XML of a WFS feature element.

    This monkeypatch requires a module variable ``location_wfs_feature``
    with the path to the wfs_feature file on disk.

    Parameters
    ----------
    request : pytest.fixture
        PyTest fixture providing request context.

    Returns
    -------
    etree.Element
        XML element representing a single record of the Boring WFS layer.

    """
    file_path = getattr(request.module, "location_wfs_feature")
    with open(file_path, 'r') as f:
        return etree.fromstring(f.read())


@pytest.fixture
def mp_remote_wfs_feature(monkeypatch, request):
    """Monkeypatch the call to get WFS features.

    This monkeypatch requires a module variable ``location_wfs_getfeature``
    with the path to the wfs_getfeature file on disk.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """
    def __get_remote_wfs_feature(*args, **kwargs):
        file_path = getattr(request.module, "location_wfs_getfeature")
        with open(file_path, 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.owsutil,
                        'wfs_get_feature',
                        __get_remote_wfs_feature)


@pytest.fixture
def mp_dov_xml(monkeypatch, request):
    """Monkeypatch the call to get the remote XML data.

    This monkeypatch requires a module variable ``location_dov_xml``
    with the path to the dov_xml file on disk.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """

    def _get_xml_data(*args, **kwargs):
        file_path = getattr(request.module, "location_dov_xml")
        with open(file_path, 'r', encoding="utf-8") as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.types.abstract.AbstractDovType,
                        '_get_xml_data', _get_xml_data)


@pytest.fixture
def mp_dov_xml_broken(monkeypatch):
    """Monkeypatch the call to break the fetching of remote XML data.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.

    """

    def _get_xml_data(*args, **kwargs):
        raise RuntimeError

    monkeypatch.setattr(pydov.types.abstract.AbstractDovType,
                        '_get_xml_data', _get_xml_data)


@pytest.fixture
def mp_geonetwork_broken(monkeypatch):
    """Monkeypatch the parsing of remote metadata to simulate the case where
    this would fail.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    """

    def _parse_remote_metadata(self, *args, **kwargs):
        for metadataUrl in self.metadataUrls:
            metadataUrl["metadata"] = None

    monkeypatch.setattr(ContentMetadata,
                        'parse_remote_metadata', _parse_remote_metadata)


@pytest.fixture
def mp_remote_xsd(monkeypatch, request):
    """Monkeypatch the call to get the remote XSD schemas.

    This monkeypatch requires a module variable ``location_xsd_base``
    with a glob expression to the XSD file(s) on disk.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    request : pytest.fixture
        PyTest fixture providing request context.

    """

    def _get_remote_xsd(*args, **kwargs):
        xsd_base_path = getattr(request.module, "location_xsd_base")
        schemas = []

        for xsd_file in glob.glob(xsd_base_path):
            with open(xsd_file, 'r', encoding="utf-8") as f:
                data = f.read()
                if not isinstance(data, bytes):
                    data = data.encode('utf-8')
                schemas.append(etree.fromstring(data))

        return schemas

    monkeypatch.setattr(pydov.search.abstract.AbstractSearch,
                        '_get_remote_xsd_schemas', _get_remote_xsd)


@pytest.fixture
def md_metadata(wfs, mp_remote_md):
    """PyTest fixture providing a MD_Metadata instance of the
    dov-pub:Boringen layer.

    Parameters
    ----------
    wfs : pytest.fixture returning owslib.wfs.WebFeatureService
        WebFeatureService based on the local GetCapabilities.
    mp_remote_md : pytest.fixture
        Monkeypatch the call to get the remote metadata of the
        dov-pub:Boringen layer.

    Returns
    -------
    owslib.iso.MD_Metadata
        Parsed metadata describing the Boringen WFS layer in more detail,
        in the ISO 19115/19139 format.

    """
    contentmetadata = wfs.contents['dov-pub:Boringen']
    return owsutil.get_remote_metadata(contentmetadata)


@pytest.fixture
def mp_remote_xml(monkeypatch):
    """Monkeypatch the call to get the remote Boring XML data.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.

    """

    def _get_remote_data(*args, **kwargs):
        with open('tests/data/types/boring/boring.xml', 'r') as f:
            data = f.read()
            if not isinstance(data, bytes):
                data = data.encode('utf-8')
        return data

    monkeypatch.setattr(pydov.util.caching.AbstractFileCache,
                        '_get_remote', _get_remote_data)


@pytest.fixture
def mp_gml_id(monkeypatch):
    """Monkeypatch the call to generate a predictable GML id.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.
    """
    def _get_id(*args, **kwargs):
        return 'pydov.gmlid'

    monkeypatch.setattr(pydov.util.location.AbstractLocation,
                        '_get_id', _get_id)


@pytest.fixture
def plaintext_cache(request):
    """Fixture for a temporary cache.

    This fixture should be parametrized, with a list of parameters in the
    order described below.

    Parameters
    ----------
    request : pytest.fixture
        PyTest fixture providing request context.

    Fixture parameters
    ------------------
    max_age : datetime.timedelta
        The maximum age to use for the cache.

    """
    orig_cache = pydov.cache

    if len(request.param) == 0:
        max_age = datetime.timedelta(seconds=1)
    else:
        max_age = request.param[0]

    plaintext_cache = PlainTextFileCache(
        cachedir=os.path.join(tempfile.gettempdir(), 'pydov_tests'),
        max_age=max_age)
    pydov.cache = plaintext_cache

    yield plaintext_cache

    plaintext_cache.remove()
    pydov.cache = orig_cache


@pytest.fixture
def gziptext_cache(request):
    """Fixture for a temporary cache.

    This fixture should be parametrized, with a list of parameters in the
    order described below.

    Parameters
    ----------
    request : pytest.fixture
        PyTest fixture providing request context.

    Fixture parameters
    ------------------
    max_age : datetime.timedelta
        The maximum age to use for the cache.

    """
    orig_cache = pydov.cache

    if len(request.param) == 0:
        max_age = datetime.timedelta(seconds=1)
    else:
        max_age = request.param[0]

    gziptext_cache = GzipTextFileCache(
        cachedir=os.path.join(tempfile.gettempdir(), 'pydov_tests'),
        max_age=max_age)
    pydov.cache = gziptext_cache

    yield gziptext_cache

    gziptext_cache.remove()
    pydov.cache = orig_cache


@pytest.fixture
def nocache():
    """Fixture to temporarily disable caching."""
    orig_cache = pydov.cache
    pydov.cache = None
    yield
    pydov.cache = orig_cache


@pytest.fixture(autouse=True)
def patch_owslib_openURL(monkeypatch):
    """Fixture to patch OWSLib's openURL function in favor of a GET request
    using our pydov requests session.

    Parameters
    ----------
    monkeypatch : pytest.fixture
        PyTest monkeypatch fixture.

    """
    def _openURL(*args, **kwargs):
        """Patch function for owslib.util.openURL using our custom pydov
        requests session.

        Parameters
        ----------
        url : str
            URL to open.

        Returns
        -------
        ResponseWrapper
            Wrapped response of the request.
        """
        url = args[0]
        return ResponseWrapper(pydov.session.get(url))

    monkeypatch.setattr('owslib.util.openURL', _openURL)
    monkeypatch.setattr('owslib.feature.common.openURL', _openURL)
    monkeypatch.setattr('owslib.feature.schema.openURL', _openURL)
    monkeypatch.setattr('owslib.feature.wfs110.openURL', _openURL)



"""Module grouping tests for the pydov.types.bodemdiepteinterval module."""

from pydov.types.bodemdiepteinterval import Bodemdiepteinterval
from pydov.types.fields import ReturnFieldList
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/bodemdiepteinterval/' \
    'wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemdiepteinterval/feature.xml'


class TestBodemdiepteinterval(AbstractTestTypes):
    """Class grouping tests for the 
    pydov.types.bodemdiepteinterval.Bodemdiepteinterval class."""

    datatype_class = Bodemdiepteinterval
    namespace = 'https://www.dov.vlaanderen.be/bodem'
    pkey_base = build_dov_url('data/bodemdiepteinterval/')

    field_names = [
        'pkey_diepteinterval', 'pkey_bodemopbouw', 'pkey_bodemlocatie',
        'nr', 'type', 'naam', 'bovengrens1_cm', 'bovengrens2_cm',
        'ondergrens1_cm', 'ondergrens2_cm', 'ondergrens_bereikt',
        'grensduidelijkheid', 'grensregelmatigheid', 'beschrijving', 'x', 'y',
        'mv_mtaw']
    field_names_subtypes = []
    field_names_nosubtypes = [
        'pkey_diepteinterval', 'pkey_bodemopbouw', 'pkey_bodemlocatie',
        'nr', 'type', 'naam', 'bovengrens1_cm', 'bovengrens2_cm',
        'ondergrens1_cm', 'ondergrens2_cm', 'ondergrens_bereikt',
        'grensduidelijkheid', 'grensregelmatigheid', 'beschrijving', 'x', 'y',
        'mv_mtaw']

    valid_returnfields = ReturnFieldList.from_field_names('pkey_diepteinterval', 'naam')
    valid_returnfields_subtype = None

    inexistent_field = 'onbestaand'

    def test_get_field_names_wrongreturnfields_nosubtypes(self):
        """Test the get_field_names method when disabling subtypes
        and including an otherwise valid return field.

        Test whether an InvalidFieldError is raised.

        Override because bodemdiepteinterval has no subtypes.
        """
        assert True



"""Module grouping tests for the generic search module, using fixed testdata."""

from pydov.search.generic import WfsSearch

location_md_metadata = \
    'tests/data/types/generic/fixed/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/generic/fixed/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/generic/fixed/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/generic/fixed/' \
                          'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/generic/fixed/feature.xml'
location_dov_xml = None
location_xsd_base = 'tests/data/types/generic/fixed/xsd_*.xml'


class TestWfsSearchFixed:
    def test_get_fields_missing_from_fc(
            self, mp_wfs, mp_get_schema,
            mp_remote_describefeaturetype, mp_remote_md,
            mp_remote_fc, mp_remote_xsd):
        """Test whether fields that are missing from the feature catalogue are
        still included in the get_fields."""

        s = WfsSearch('pfas:pfas_analyseresultaten')
        fields = s.get_fields()
        assert 'x_ml72' in fields

    def test_output_fields_missing_from_fc(
            self, mp_wfs, mp_get_schema, mp_remote_describefeaturetype,
            mp_remote_md, mp_remote_fc, mp_remote_wfs_feature,
            mp_dov_xml):
        """Test whether fields that are missing from the feature catalogue are
        still included in the output."""

        s = WfsSearch('pfas:pfas_analyseresultaten')
        df = s.search(max_features=1)
        assert 'x_ml72' in list(df)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV
bodemdiepteinterval data."""
from ..types.bodemdiepteinterval import Bodemdiepteinterval
from .abstract import AbstractSearch


class BodemdiepteintervalSearch(AbstractSearch):
    """Search class to retrieve information about bodemdiepteintervallen."""

    def __init__(self, objecttype=Bodemdiepteinterval):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemdiepteinterval type.
            Optional: defaults to the Bodemdiepteinterval type containing the
            fields described in the documentation.

        """
        super(BodemdiepteintervalSearch, self).__init__(
            'bodem:bodemdiepteintervallen', objecttype)



"""Module grouping tests for the pydov.types.generic module."""

from pydov.types.abstract import AbstractDovType
from pydov.types.generic import WfsTypeFactory
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/generic/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/generic/feature.xml'
location_dov_xml = None


class TestWfsTypeFactory():
    """Class grouping tests for the pydov.types.generic.WfsTypeFactory class."""

    def test_get_wfs_type(self):
        """Test creating a WfsType for a given test layer.

        Test whether the created type is a subclass of AbstractDovType.
        """
        erfgoedType = WfsTypeFactory.get_wfs_type(
            'bodem_varia:bodemkundig_erfgoed')
        assert issubclass(erfgoedType, AbstractDovType)


class TestWfsTypeFactoryOpdracht(AbstractTestTypes):
    """Class grouping tests for the pydov.types.generic.WfsTypeFactory class
    for the dov-pub:Opdrachten type."""

    datatype_class = WfsTypeFactory.get_wfs_type('dov-pub:Opdrachten')
    namespace = 'http://dov.vlaanderen.be/dov-pub/Opdrachten'
    pkey_base = None

    field_names = []
    field_names_subtypes = None
    field_names_nosubtypes = []

    valid_returnfields = None
    valid_returnfields_subtype = None

    inexistent_field = 'onbestaand'

    def test_get_field_names_returnfields_nosubtypes(self):
        """WfsType have no fixed fields."""
        assert True

    def test_get_field_names_returnfields_order(self):
        """WfsType have no fixed fields."""
        assert True

    def test_get_field_names_wrongreturnfields(self):
        """WfsType have no fixed fields."""
        assert True

    def test_get_field_names_wrongreturnfieldstype(self):
        """WfsType have no fixed fields."""
        assert True



"""Module grouping tests for the pydov.util.owsutil module."""
import copy

import pytest
from owslib.etree import etree
from owslib.fes2 import FilterRequest, PropertyIsEqualTo, SortBy, SortProperty, Or
from owslib.iso import MD_Metadata
from owslib.util import nspath_eval

from pydov.util import owsutil
from pydov.util.dovutil import build_dov_url
from pydov.util.location import Box, Within, WithinDistance
from tests.abstract import clean_xml

location_md_metadata = 'tests/data/types/boring/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/boring/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/boring/wfsdescribefeaturetype.xml'


class TestOwsutil(object):
    """Class grouping tests for the pydov.util.owsutil module."""

    def test_get_csw_base_url(self, wfs):
        """Test the owsutil.get_csw_base_url method.

        Test whether the CSW base URL of the dov-pub:Boringen layer is correct.

        Parameters
        ----------
        wfs : pytest.fixture returning owslib.wfs.WebFeatureService
            WebFeatureService based on the local GetCapabilities.

        """
        contentmetadata = wfs.contents['dov-pub:Boringen']
        assert owsutil.get_csw_base_url(contentmetadata) == \
            build_dov_url('geonetwork/srv/dut/csw')

    def test_get_csw_base_url_nometadataurls(self, wfs):
        """Test the owsutil.get_csw_base_url method for a layer without
        metadata urls.

        Test whether None is returned.

        Parameters
        ----------
        wfs : pytest.fixture returning owslib.wfs.WebFeatureService
            WebFeatureService based on the local GetCapabilities.

        """
        contents = copy.deepcopy(wfs.contents)
        contentmetadata = contents['dov-pub:Boringen']
        contentmetadata.metadataUrls = []

        assert owsutil.get_csw_base_url(contentmetadata) is None

    def test_get_featurecatalogue_uuid(self, md_metadata):
        """Test the owsutil.get_featurecatalogue_uuid method.

        Test whether the featurecatalogue uuid of the dov-pub:Boringen layer
        is correct.

        Parameters
        ----------
        md_metadata : pytest.fixture providing owslib.iso.MD_Metadata
            Parsed metadata describing the Boringen WFS layer in more detail,
            in the ISO 19115/19139 format.

        """
        assert owsutil.get_featurecatalogue_uuid(md_metadata) == \
            'c0cbd397-520f-4ee1-aca7-d70e271eeed6'

    def test_get_featurecatalogue_uuid_nocontentinfo(self, md_metadata):
        """Test the owsutil.get_featurecatalogue_uuid method when the
        metadata is missing the gmd:contentInfo element.

        Test whether None is returned.

        Parameters
        ----------
        md_metadata : pytest.fixture providing owslib.iso.MD_Metadata
            Parsed metadata describing the Boringen WFS layer in more detail,
            in the ISO 19115/19139 format.

        """
        tree = etree.fromstring(md_metadata.xml)
        for ci in tree.findall(
                './/{http://www.isotc211.org/2005/gmd}contentInfo'):
            tree.remove(ci)
        md_metadata = MD_Metadata(tree)

        assert owsutil.get_featurecatalogue_uuid(md_metadata) is None

    def test_get_featurecatalogue_uuid_nouuidref(self, md_metadata):
        """Test the owsutil.get_featurecatalogue_uuid method when the
        gmd:contentInfo element is missing a 'uuidref' attribute.

        Test whether None is returned.

        Parameters
        ----------
        md_metadata : pytest.fixture providing owslib.iso.MD_Metadata
            Parsed metadata describing the Boringen WFS layer in more detail,
            in the ISO 19115/19139 format.

        """
        tree = etree.fromstring(md_metadata.xml)
        for ci in tree.findall(nspath_eval(
            'gmd:contentInfo/'
            'gmd:MD_FeatureCatalogueDescription/'
            'gmd:featureCatalogueCitation',
                {'gmd': 'http://www.isotc211.org/2005/gmd'})):
            ci.attrib.pop('uuidref')
        md_metadata = MD_Metadata(tree)

        assert owsutil.get_featurecatalogue_uuid(md_metadata) is None

    def test_get_namespace(self, wfs, mp_remote_describefeaturetype):
        """Test the owsutil.get_namespace method.

        Test whether the namespace of the dov-pub:Boringen layer is correct.

        Parameters
        ----------
        wfs : pytest.fixture returning owslib.wfs.WebFeatureService
            WebFeatureService based on the local GetCapabilities.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType of the
            dov-pub:Boringen layer.

        """
        assert owsutil.get_namespace(wfs, 'dov-pub:Boringen') == \
            'http://dov.vlaanderen.be/ocdov/dov-pub'

    def test_get_remote_featurecatalogue(self, mp_remote_fc):
        """Test the owsutil.get_remote_featurecatalogue method.

        Test whether the feature catalogue of the dov-pub:Boringen layer
        matches the format described in the docs.

        Parameters
        ----------
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue of the
            dov-pub:Boringen layer.

        """
        fc = owsutil.get_remote_featurecatalogue(
            build_dov_url('geonetwork/srv/nl/csw'),
            'c0cbd397-520f-4ee1-aca7-d70e271eeed6')

        assert isinstance(fc, dict)

        assert 'definition' in fc
        assert isinstance(fc['definition'], str)

        assert 'attributes' in fc
        assert isinstance(fc['attributes'], dict)

        attrs = fc['attributes']
        if len(attrs) > 0:
            for attr in attrs.values():
                assert isinstance(attr, dict)

                assert 'definition' in attr
                assert isinstance(attr['definition'], str)

                assert 'values' in attr

                if attr['values'] is not None:
                    assert isinstance(attr['values'], dict)

                    for v in attr['values'].keys():
                        assert isinstance(v, str)
                        assert isinstance(attr['values'][v], str) or \
                            attr['values'][v] is None
                    assert len(attr['values'].keys()) == len(
                        set(attr['values'].keys()))

                assert 'multiplicity' in attr
                mp = attr['multiplicity']
                assert isinstance(mp, tuple)
                assert len(mp) == 2
                assert mp[0] in (0, 1)
                assert (isinstance(mp[1], int) and mp[1] > 0) or mp[1] == 'Inf'

    def test_get_remote_featurecataloge_baduuid(self, mp_remote_fc_notfound):
        """Test the owsutil.get_remote_featurecatalogue method with an
        inexistent feature catalogue uuid.

        Test whether None is returned.

        Parameters
        ----------
        mp_remote_fc_notfound : pytest.fixture
            Monkeypatch the call to get an inexistent remote featurecatalogue.

        """
        assert owsutil.get_remote_featurecatalogue(
            build_dov_url('geonetwork/srv/nl/csw'),
            'badfc000-0000-0000-0000-badfc00badfc') is None

    def test_get_remote_metadata(self, md_metadata):
        """Test the owsutil.get_remote_metadata method.

        Test whether the resulting MD_Metadata is correct.

        Parameters
        ----------
        md_metadata : pytest.fixture returning owslib.iso.MD_Metadata
            Parsed metadata describing the Boringen WFS layer in more detail,
            in the ISO 19115/19139 format.

        """
        assert isinstance(md_metadata, MD_Metadata)

    def test_get_wfs_max_features(self, wfs_capabilities):
        """Test the owsutil.get_wfs_max_features method.

        Test whether the default maximum number of features can be found.

        Parameters
        ----------
        wfs_capabilities : pytest.fixture
            PyTest fixture providing the WFS GetCapabilities response.
        """
        max_features = owsutil.get_wfs_max_features(wfs_capabilities)

        assert isinstance(max_features, int)
        assert max_features > 0


class TestWfsGetFeatureRequest(object):
    """Class grouping tests checking the WFS GetFeature request generation."""

    def test_wfs_build_getfeature_request_onlytypename(self):
        """Test the owsutil.wfs_build_getfeature_request method with only a
        typename specified.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request('dov-pub:Boringen')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"><wfs:Query '
            'typeNames="dov-pub:Boringen"/></wfs'
            ':GetFeature>')

    def test_wfs_build_getfeature_start_index(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        startIndex defined.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', start_index=3)

        assert "startIndex" in xml.attrib.keys()
        assert xml.attrib["startIndex"] == "3"

    def test_wfs_build_getfeature_start_index_negative(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a negative startIndex value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', start_index=-5)

    def test_wfs_build_getfeature_start_index_none(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a startIndex value of None.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', start_index=None)

    def test_wfs_build_getfeature_start_index_float(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a floating point startIndex value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', start_index=1.5)

    def test_wfs_build_getfeature_start_index_string(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        an non-integer startIndex value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', start_index="0")

    def test_wfs_build_getfeature_maxfeatures(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        limited set of features defined.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', max_features=3)

        assert "count" in xml.attrib.keys()
        assert xml.attrib["count"] == "3"

    def test_wfs_build_getfeature_maxfeatures_negative(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a negative maxfeature value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', max_features=-5)

    def test_wfs_build_getfeature_maxfeatures_float(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a floating point maxfeature value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', max_features=1.5)

    def test_wfs_build_getfeature_maxfeatures_zero(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        a maxfeature value of 0.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', max_features=0)

    def test_wfs_build_getfeature_maxfeatures_string(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        an non-integer maxfeature value.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen', max_features="0")

    def test_wfs_build_getfeature_request_bbox_nogeometrycolumn(self):
        """Test the owsutil.wfs_build_getfeature_request method with a location
        argument but without the geometry_column argument.

        Test whether an AttributeError is raised.

        """
        with pytest.raises(AttributeError):
            owsutil.wfs_build_getfeature_request(
                'dov-pub:Boringen',
                location=Within(Box(151650, 214675, 151750, 214775)))

    def test_wfs_build_getfeature_request_bbox(self, mp_gml_id):
        """Test the owsutil.wfs_build_getfeature_request method with a
        typename, box and geometry_column.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen',
            location=Within(Box(151650, 214675, 151750, 214775)),
            geometry_column='geom')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"><wfs:Query '
            'typeNames="dov-pub:Boringen"><fes:Filter '
            'xmlns:fes="http://www.opengis.net/fes/2.0"><fes:Within> '
            '<fes:ValueReference>geom</fes:ValueReference><gml:Envelope '
            'xmlns:gml="http://www.opengis.net/gml/3.2" srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370" gml32:id='
            '"pydov.gmlid"><gml:lowerCorner>151650.000000 '
            '214675.000000</gml:lowerCorner><gml:upperCorner>151750.000000 '
            '214775.000000</gml:upperCorner></gml:Envelope></fes:Within></fes'
            ':Filter></wfs:Query></wfs:GetFeature>')

    def test_wfs_build_getfeature_request_gml_id_stable(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        typename, box and geometry_column.
        Test whether the XML of the WFS GetFeature call is stable.
        """
        xml1 = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen',
            location=Within(Box(151650, 214675, 151750, 214775)),
            geometry_column='geom')

        xml2 = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen',
            location=Within(Box(151650, 214675, 151750, 214775)),
            geometry_column='geom')

        assert etree.tostring(xml1) == etree.tostring(xml2)

    def test_wfs_build_getfeature_request_gml_id_unique(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        typename, two boxes and geometry_column.
        Test whether the GML ids in the XML of the WFS GetFeature are unique.
        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen',
            location=Or([
                Within(Box(100000, 120000, 200000, 220000)),
                WithinDistance(Box(100000, 120000, 200000, 220000), 10)
            ]),
            geometry_column='geom')

        gml_items = xml.findall('.//*[@{http://www.opengis.net/gml/3.2}id]')
        gml_ids = [i.get('{http://www.opengis.net/gml/3.2}id') for i in gml_items]

        assert len(gml_ids) == len(set(gml_ids))

    def test_wfs_build_getfeature_request_propertyname(self):
        """Test the owsutil.wfs_build_getfeature_request method with a list
        of propertynames.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', propertyname=['fiche', 'diepte_tot_m'])

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"> <wfs:Query '
            'typeNames="dov-pub:Boringen"> '
            '<wfs:PropertyName>diepte_tot_m</wfs:PropertyName> '
            '<wfs:PropertyName>fiche</wfs:PropertyName>'
            '</wfs:Query> </wfs:GetFeature>')

    def test_wfs_build_getfeature_request_propertyname_stable(self):
        """Test the owsutil.wfs_build_getfeature_request method with a list
        of propertynames.

        Test whether the XML of the WFS GetFeature that is being generated is
        stable (i.e. independent of the order of the propertynames).

        """
        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', propertyname=['fiche', 'diepte_tot_m'])

        xml2 = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', propertyname=['diepte_tot_m', 'fiche'])

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            etree.tostring(xml2).decode('utf8'))

    def test_wfs_build_getfeature_request_filter(self):
        """Test the owsutil.wfs_build_getfeature_request method with an
        attribute filter.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        query = PropertyIsEqualTo(propertyname='gemeente',
                                  literal='Herstappe')
        filter_request = FilterRequest()
        filter_request = filter_request.setConstraint(query)
        filter_request = etree.tostring(filter_request, encoding='unicode')

        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', filter=filter_request)

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"> <wfs:Query '
            'typeNames="dov-pub:Boringen"> <fes:Filter> '
            '<fes:PropertyIsEqualTo> '
            '<fes:ValueReference>gemeente</fes:ValueReference> '
            '<ogc:Literal>Herstappe</ogc:Literal> </fes:PropertyIsEqualTo> '
            '</ogc:Filter> </wfs:Query> </wfs:GetFeature>')

    def test_wfs_build_getfeature_request_bbox_filter(self, mp_gml_id):
        """Test the owsutil.wfs_build_getfeature_request method with an
        attribute filter, a box and a geometry_column.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        query = PropertyIsEqualTo(propertyname='gemeente',
                                  literal='Herstappe')
        filter_request = FilterRequest()
        filter_request = filter_request.setConstraint(query)
        filter_request = etree.tostring(filter_request, encoding='unicode')

        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', filter=filter_request,
            location=Within(Box(151650, 214675, 151750, 214775)),
            geometry_column='geom')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"> <wfs:Query '
            'typeNames="dov-pub:Boringen"> <fes:Filter> <fes:And> '
            '<fes:PropertyIsEqualTo> '
            '<fes:ValueReference>gemeente</fes:ValueReference> '
            '<ogc:Literal>Herstappe</ogc:Literal> </fes:PropertyIsEqualTo> '
            '<fes:Within> <fes:ValueReference>geom</fes:ValueReference> '
            '<gml:Envelope xmlns:gml="http://www.opengis.net/gml" '
            'srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370" gml32:id='
            '"pydov.gmlid"> '
            '<gml:lowerCorner>151650.000000 214675.000000</gml:lowerCorner> '
            '<gml:upperCorner>151750.000000 214775.000000</gml:upperCorner> '
            '</gml:Envelope> </fes:Within> </fes:And> </fes:Filter> '
            '</wfs:Query> </wfs:GetFeature>')

    def test_wfs_build_getfeature_request_bbox_filter_propertyname(
            self, mp_gml_id):
        """Test the owsutil.wfs_build_getfeature_request method with an
        attribute filter, a box, a geometry_column and a list of
        propertynames.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        query = PropertyIsEqualTo(propertyname='gemeente',
                                  literal='Herstappe')
        filter_request = FilterRequest()
        filter_request = filter_request.setConstraint(query)
        filter_request = etree.tostring(filter_request, encoding='unicode')

        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', filter=filter_request,
            location=Within(Box(151650, 214675, 151750, 214775)),
            geometry_column='geom', propertyname=['fiche', 'diepte_tot_m'])

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"> <wfs:Query '
            'typeNames="dov-pub:Boringen"> '
            '<wfs:PropertyName>diepte_tot_m</wfs:PropertyName> '
            '<wfs:PropertyName>fiche</wfs:PropertyName> <fes:Filter> '
            '<fes:And> <fes:PropertyIsEqualTo> '
            '<fes:ValueReference>gemeente</fes:ValueReference> '
            '<ogc:Literal>Herstappe</ogc:Literal> </fes:PropertyIsEqualTo> '
            '<fes:Within> <fes:ValueReference>geom</fes:ValueReference> '
            '<gml:Envelope xmlns:gml="http://www.opengis.net/gml" '
            'srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370" gml32:id='
            '"pydov.gmlid"> '
            '<gml:lowerCorner>151650.000000 214675.000000</gml:lowerCorner> '
            '<gml:upperCorner>151750.000000 214775.000000</gml:upperCorner> '
            '</gml:Envelope> </fes:Within> </fes:And> </fes:Filter> '
            '</wfs:Query> </wfs:GetFeature>')

    def test_wfs_build_getfeature_request_sortby(self):
        """Test the owsutil.wfs_build_getfeature_request method with a sortby.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        sort_by = SortBy([SortProperty('diepte_tot_m', 'DESC')])

        sort_by = etree.tostring(sort_by.toXML(), encoding='unicode')

        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', propertyname=['fiche', 'diepte_tot_m'],
            sort_by=sort_by)

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"><wfs:Query '
            'typeNames="dov-pub:Boringen"><wfs:PropertyName>diepte_tot_m</wfs'
            ':PropertyName><wfs:PropertyName>fiche</wfs:PropertyName'
            '><fes:SortBy><fes:SortProperty><fes:ValueReference'
            '>diepte_tot_m</fes:ValueReference><ogfesc:SortOrder>DESC</fes'
            ':SortOrder></fes:SortProperty></fes:SortBy></wfs:Query></wfs'
            ':GetFeature>')

    def test_wfs_build_getfeature_request_sortby_multi(self):
        """Test the owsutil.wfs_build_getfeature_request method with a
        sortby containing multiple sort properties.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        sort_by = SortBy([SortProperty('diepte_tot_m', 'DESC'),
                          SortProperty('datum_aanvang', 'ASC')])

        sort_by = etree.tostring(sort_by.toXML(), encoding='unicode')

        xml = owsutil.wfs_build_getfeature_request(
            'dov-pub:Boringen', propertyname=['fiche', 'diepte_tot_m'],
            sort_by=sort_by)

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"><wfs:Query '
            'typeNames="dov-pub:Boringen"><wfs:PropertyName>diepte_tot_m</wfs'
            ':PropertyName><wfs:PropertyName>fiche</wfs:PropertyName'
            '><fes:SortBy><fes:SortProperty><fes:ValueReference'
            '>diepte_tot_m</fes:ValueReference><fes:SortOrder>DESC</fes'
            ':SortOrder></fes:SortProperty><fes:SortProperty><fes'
            ':ValueReference>datum_aanvang</fes:ValueReference><fes:SortOrder>'
            'ASC</fes:SortOrder></fes:SortProperty></fes:SortBy></wfs:Query>'
            '</wfs:GetFeature>')

    def test_wfs_build_getfeature_request_srs(self):
        """Test the owsutil.wfs_build_getfeature_request method with only a
        typename and an SRS.

        Test whether the XML of the WFS GetFeature call is generated correctly.

        """
        xml = owsutil.wfs_build_getfeature_request('dov-pub:Boringen', crs='EPSG:31370')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs/2.0" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
            'service="WFS" version="2.0.0" startIndex="0" '
            'xsi:schemaLocation="http://www.opengis.net/wfs/2.0 '
            'http://schemas.opengis.net/wfs/2.0/wfs.xsd"><wfs:Query '
            'typeNames="dov-pub:Boringen" srsName="EPSG:31370"/></wfs'
            ':GetFeature>')

    def test_wfs_build_getfeature_request_srs_wrongtype(self):
        """Test the owsutil.wfs_build_getfeature_request method with only a
        typename and an CRS of the wrong type.

        Test whether a TypeError is raised.

        """
        with pytest.raises(TypeError):
            owsutil.wfs_build_getfeature_request('dov-pub:Boringen', crs=31370)

    def test_wfs_build_getfeature_request_srs_wrongvalue(self):
        """Test the owsutil.wfs_build_getfeature_request method with only a
        typename and an CRS of the wrong type.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            owsutil.wfs_build_getfeature_request('dov-pub:Boringen', crs='31370')



# -*- coding: utf-8 -*-
"""Module containing the generic DOV data types."""

from pydov.types.abstract import AbstractDovType


class WfsTypeFactory:
    """Class to generate pydov type classes at runtime for given WFS layers."""

    @staticmethod
    def get_wfs_type(layer):
        """Generate a new pydov WfsType class for the given layer.

        Parameters
        ----------
        layer : str
            Workspace qualified layer name, uniquely identifying the layer
            within the DOV WFS service.

        Returns
        -------
        class
            WfsType class for the given layer.
        """
        class WfsType(AbstractDovType):

            fields = []

            def __init__(self, pkey):
                super().__init__(layer, pkey)

        return WfsType



import json
import re
import shutil
import time
import zipfile
import pytest
import tempfile
import os
import pydov

from pydov.search.boring import BoringSearch
from pydov.util.errors import LogReplayError
from pydov.util.hooks import Hooks, RepeatableLogRecorder, RepeatableLogReplayer
from tests.abstract import ServiceCheck


@pytest.fixture()
def temp_directory():
    """PyTest fixture providing a temporary directory that does not exist on
    disk.

    Yields
    ------
    str
        Path to temporary directory that does not exist on disk yet.
    """
    temp_dir = os.path.join(tempfile.gettempdir(), 'pydov-test-repeatable-log')
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir, ignore_errors=True)

    yield temp_dir

    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir, ignore_errors=True)


@pytest.fixture()
def log_recorder(temp_directory):
    """PyTest fixture providing an instance of the RepeatableLogRecorder hook,
    using a temporary directory to store the archive.

    Parameters
    ----------
    temp_directory : pytest.fixture
        Fixture providing a temporary directory.

    Yields
    ------
    RepeatableLogRecorder
        RepeatableLogRecorder hook instance.
    """
    orig_hooks = Hooks(pydov.hooks)
    recorder = RepeatableLogRecorder(temp_directory)
    pydov.hooks = Hooks((recorder,))

    yield recorder

    pydov.hooks = orig_hooks


@pytest.fixture()
def log_archive(log_recorder):
    """PyTest fixture installing the log recorder and executing a Boring search
    to record the session in an archive.

    Yields the resulting archive as ZipFile with read possibilities, for
    inspection.

    Parameters
    ----------
    log_recorder : pytest.fixture
        Fixture providing an instance of a RepeatableLogRecorder.

    Yields
    ------
    zipfile.ZipFile
        Resulting archive open in read mode.
    """
    bs = BoringSearch()
    bs.search(max_features=1)

    log_recorder._pydov_exit()

    with zipfile.ZipFile(log_recorder.log_archive, 'r') as z:
        yield z


class TestRepeatableLogRecorder(object):
    """Class grouping tests for the RepeatableLogRecorder."""

    def test_create_log_directory(self, temp_directory):
        """Test whether to log directory is created if it does not exist yet.

        Parameters
        ----------
        temp_directory : str
            Fixture providing a temporary directory.
        """
        assert not os.path.exists(temp_directory)
        RepeatableLogRecorder(temp_directory)
        assert os.path.exists(temp_directory)

    def test_create_log_archive(self, temp_directory):
        """Test whether the log archive is recorded and the filename matches the expectations.

        Parameters
        ----------
        temp_directory : str
            Fixture providing a temporary directory.
        """
        RepeatableLogRecorder(temp_directory)
        assert len(os.listdir(temp_directory)) >= 1

        log_file = os.listdir(temp_directory)[0]
        assert re.match(
            r'^pydov-archive-[0-9]{8}T[0-9]{6}-([0-9a-z]){6}.zip', log_file)

    def test_pydov_code(self, log_recorder):
        """Test whether the pydov code is recorded in the archive.

        Parameters
        ----------
        log_recorder : RepeatableLogRecorder
            Fixture providing an instance of a RepeatableLogRecorder.
        """
        log_recorder._pydov_exit()

        with zipfile.ZipFile(log_recorder.log_archive, 'r') as z:
            assert 'pydov/__init__.py' in [f.filename for f in z.filelist]

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_contents(self, log_archive):
        """Test whether the contents of the archive match the recorded session.

        Parameters
        ----------
        log_archive : zipfile.ZipFile
            Fixture providing the recorded session as archive in read mode.
        """
        assert 'metadata.json' in [f.filename for f in log_archive.filelist]

        metadata = json.loads(log_archive.read('metadata.json'))

        for lib in ['pydov', 'owslib', 'pandas', 'numpy', 'requests',
                    'fiona', 'geopandas']:
            assert lib in metadata['versions']

        for t in ['start', 'end']:
            assert t in metadata['timings']
            assert re.match(r'[0-9]{8}T[0-9]{6}', metadata['timings'][t])

        assert 'run_time_secs' in metadata['timings']

        # metadata
        assert len([f.filename for f in log_archive.filelist
                    if f.filename.startswith('meta/')]) == 2

        # wfs getfeature
        assert len([f.filename for f in log_archive.filelist
                    if f.filename.startswith('wfs/')]) == 1

        # DOV xml
        assert len([f.filename for f in log_archive.filelist
                    if f.filename.startswith('xml/')]) == 1


class TestRepeatableLogReplayer(object):
    """Class grouping tests for the RepeatableLogReplayer."""
    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_replay(self, log_recorder):
        """Test whether a recorded session can be replayed using the archive.

        Parameters
        ----------
        log_recorder : RepeatableLogRecorder
            Fixture providing an instance of a RepeatableLogRecorder.
        """
        bs = BoringSearch()
        bs.search(max_features=1)
        log_recorder._pydov_exit()

        zip_contents = dict()

        with zipfile.ZipFile(log_recorder.log_archive, 'r') as z:
            for f in [f.filename for f in z.filelist]:
                zip_contents[f] = z.read(f)

        with zipfile.ZipFile(
                log_recorder.log_archive, 'w',
                compression=zipfile.ZIP_DEFLATED) as z:

            for f in zip_contents:
                if f.startswith('wfs/'):
                    z.writestr(f, zip_contents[f].decode('utf-8').replace(
                        'gemeente>Wortegem-Petegem<',
                        'gemeente>Wortelgem<').encode('utf-8')
                    )
                else:
                    z.writestr(f, zip_contents[f])

        pydov.hooks = Hooks((RepeatableLogReplayer(log_recorder.log_archive),))

        bs = BoringSearch()
        df = bs.search(max_features=1)
        assert df.iloc[0].gemeente == 'Wortelgem'

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_replay_other_fails(self, log_recorder):
        """Test whether an error is raised when a session is modified and
        cannot be replayed using the saved archive.

        Parameters
        ----------
        log_recorder : RepeatableLogRecorder
            Fixture providing an instance of a RepeatableLogRecorder.
        """
        bs = BoringSearch()
        bs.search(max_features=1)
        log_recorder._pydov_exit()

        pydov.hooks = Hooks((RepeatableLogReplayer(log_recorder.log_archive),))

        with pytest.raises(LogReplayError):
            bs = BoringSearch()
            bs.search(max_features=2)



# -*- coding: utf-8 -*-
"""Module containing generic fields for the location (ligging) of objects."""

import numpy as np

from pydov.types.fields import _CustomXmlField


class MvMtawField(_CustomXmlField):
    """Field for retrieving the mv_mtaw value from the height of the point in
    relevant cases."""

    def __init__(self, definition):
        """Initialise a MvMtawField (mv_mtaw) with given definition.

        Parameters
        ----------
        definition : string
            Type-specific definition of the mv_mtaw field.
        """
        super().__init__(
            name='mv_mtaw',
            definition=definition,
            datatype='float',
            notnull=False
        )

    def calculate(self, cls, tree):
        # Support the old format too
        oorspronkelijk_maaiveld = cls._parse(
            func=tree.findtext,
            xpath='.//oorspronkelijk_maaiveld/waarde',
            namespace=None,
            returntype='float'
        )
        if oorspronkelijk_maaiveld is not np.nan:
            return oorspronkelijk_maaiveld

        # Check if referentiepunt is Maaiveld
        referentiepunt = cls._parse(
            func=tree.findtext,
            xpath='.//ligging/metadata_hoogtebepaling/referentiepunt_type',
            namespace=None,
            returntype='string'
        )
        if referentiepunt != 'Maaiveld':
            # If referentiepunt is not Maaiveld, we don't know the height of
            # maaiveld
            return np.nan

        # If referentiepunt is Maaiveld, height of the ligging is Maaiveld
        point = tree.findtext(
            './/ligging/{http://www.opengis.net/gml/3.2}Point/'
            '{http://www.opengis.net/gml/3.2}pos'
        )
        coords = point.split(' ')
        if len(coords) == 3:
            hoogte = float(coords[-1])
            return hoogte
        else:
            return np.nan



import pandas as pd
import pytest
from pydov.search.interpretaties import InformeleStratigrafieSearch

from tests.abstract import ServiceCheck


class TestSearchCustomWfsField(object):
    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_search_resolve_customfield(self):
        """Test the search method without explicitly requesting the
            'Type_proef' and 'Proeffiche' fields.

            Test whether the output dataframe includes values for pkey_boring or
            pkey_sondering.

            Parameters
            ----------
            mp_get_schema : pytest.fixture
                Monkeypatch the call to a remote OWSLib schema.
            mp_remote_describefeaturetype : pytest.fixture
                Monkeypatch the call to a remote DescribeFeatureType .
            mp_remote_wfs_feature : pytest.fixture
                Monkeypatch the call to get WFS features.
            mp_dov_xml : pytest.fixture
                Monkeypatch the call to get the remote XML data.

            """
        df = InformeleStratigrafieSearch().search(
            max_features=1,
            return_fields=('pkey_interpretatie', 'pkey_boring',
                           'pkey_sondering'))

        assert (not pd.isnull(df.pkey_boring[0])) or (
            not pd.isnull(df.pkey_sondering[0]))



"""Module grouping tests for the
pydov.types.interpretaties.InformeleStratigrafie class."""
from pydov.types.fields import ReturnFieldList
from pydov.types.interpretaties import HydrogeologischeStratigrafie
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = \
    'tests/data/types/interpretaties/hydrogeologische_stratigrafie/' \
    'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/interpretaties/hydrogeologische_stratigrafie/feature.xml'
location_dov_xml = \
    'tests/data/types/interpretaties/hydrogeologische_stratigrafie' \
    '/hydrogeologische_stratigrafie.xml'


class TestHydrogeologischeStratigrafie(AbstractTestTypes):
    """Class grouping tests for the
    pydov.types.interpretaties.HydrogeologischeStratigrafie class."""

    datatype_class = HydrogeologischeStratigrafie
    namespace = 'http://dov.vlaanderen.be/ocdov/interpretaties'
    pkey_base = build_dov_url('data/interpretatie/')

    field_names = [
        'pkey_interpretatie', 'pkey_boring',
        'betrouwbaarheid_interpretatie', 'x', 'y',
        'start_interpretatie_mtaw',
        'diepte_laag_van', 'diepte_laag_tot', 'aquifer']
    field_names_subtypes = [
        'diepte_laag_van', 'diepte_laag_tot', 'aquifer']
    field_names_nosubtypes = [
        'pkey_interpretatie', 'pkey_boring',
        'betrouwbaarheid_interpretatie', 'x', 'y',
        'start_interpretatie_mtaw']

    valid_returnfields = ReturnFieldList.from_field_names('pkey_interpretatie', 'pkey_boring')
    valid_returnfields_subtype = ReturnFieldList.from_field_names(
        'pkey_interpretatie', 'diepte_laag_van', 'diepte_laag_tot')

    inexistent_field = 'onbestaand'



"""Module grouping tests for the generic search module."""

from pydov.search.generic import WfsSearch
from pydov.types.fields import ReturnFieldList
from pydov.types.generic import WfsTypeFactory
from owslib.fes2 import PropertyIsEqualTo
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = \
    'tests/data/types/generic/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/generic/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/generic/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/generic/' \
                          'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/generic/feature.xml'
location_dov_xml = None
location_xsd_base = 'tests/data/types/generic/xsd_*.xml'


class TestWfsSearch(AbstractTestSearch):

    search_instance = WfsSearch('dov-pub:Opdrachten')
    datatype_class = WfsTypeFactory.get_wfs_type('dov-pub:Opdrachten')

    valid_query_single = PropertyIsEqualTo(
        propertyname='fiche',
        literal=build_dov_url('data/opdracht/2021-026141%27'))

    inexistent_field = 'onbestaand'
    xml_field = None
    wfs_field = 'opdrachtgever'

    valid_returnfields = ReturnFieldList.from_field_names('opdrachtgever', 'opdrachtnemer')
    valid_returnfields_subtype = None
    valid_returnfields_extra = ReturnFieldList.from_field_names('startdatum', 'naam')

    df_default_columns = [
        'id', 'naam', 'fiche', 'omschrijving', 'startdatum', 'einddatum',
        'opdrachtgever', 'opdrachtnemer', 'dataleverancier', 'aard',
        'origine', 'eerste_invoer']

    def test_pluggable_type(self):
        """Generic WFS types are not pluggable."""
        assert True



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for GW permits."""
from pydov.types.abstract import AbstractDovType
from pydov.types.fields import WfsField


class GrondwaterVergunning(AbstractDovType):
    """Class representing the DOV data type for groundwater abstraction
    permits."""

    subtypes = []

    fields = [
        WfsField(name='id_vergunning',
                 source_field='id',
                 datatype='string'),
        WfsField(name='pkey_installatie',
                 source_field='installatie',
                 datatype='string'),
        WfsField(name='x',
                 source_field='installatie_X_mL72',
                 datatype='float'),
        WfsField(name='y',
                 source_field='installatie_Y_mL72',
                 datatype='float'),
        WfsField(name='diepte',
                 source_field='vergunde_diepte_m',
                 datatype='float'),
        WfsField(name='exploitant_naam',
                 source_field='exploitant_naam',
                 datatype='string'),
        WfsField(name='watnr',
                 source_field='watnr',
                 datatype='string'),
        WfsField(name='vlaremrubriek',
                 source_field='vlaremrubriek',
                 datatype='string'),
        WfsField(name='vergund_jaardebiet',
                 source_field='vergund_jaardebiet',
                 datatype='float'),
        WfsField(name='vergund_dagdebiet',
                 source_field='vergund_dagdebiet',
                 datatype='float'),
        WfsField(name='van_datum_termijn',
                 source_field='van_datum_termijn',
                 datatype='date'),
        WfsField(name='tot_datum_termijn',
                 source_field='tot_datum_termijn',
                 datatype='date'),
        WfsField(name='aquifer_vergunning',
                 source_field='aquifer_vergunning',
                 datatype='string'),
        WfsField(name='inrichtingsklasse',
                 source_field='inrichtingsklasse',
                 datatype='string'),
        WfsField(name='nacebelcode',
                 source_field='IIOA_nacebelcode',
                 datatype='string'),
        WfsField(name='actie_waakgebied',
                 source_field='actie_waakgebied',
                 datatype='string'),
        WfsField(name='cbbnr',
                 source_field='exploitant_CBBnr',
                 datatype='string'),
        WfsField(name='kbonr',
                 source_field='exploitant_KBOnr',
                 datatype='string'),
    ]

    pkey_fieldname = 'id'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Gwvergunningen (permits),
            being a URI of the form
            `https://www.dov.vlaanderen.be/data/installatie/<id>`.

        """
        super(GrondwaterVergunning, self).__init__('gwvergunning', pkey)



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for boreholes (Boring), including
subtypes."""
from pydov.types.fields import WfsField, XmlField
from pydov.types.ligging import MvMtawField

from .abstract import AbstractDovSubType, AbstractDovType


class BoorMethode(AbstractDovSubType):

    rootpath = './/boring/details/boormethode'

    fields = [
        XmlField(name='diepte_methode_van',
                 source_xpath='/van',
                 definition='Bovenkant van de laag die met een bepaalde '
                            'methode aangeboord werd, in meter.',
                 datatype='float'),
        XmlField(name='diepte_methode_tot',
                 source_xpath='/tot',
                 definition='Onderkant van de laag die met een bepaalde '
                            'methode aangeboord werd, in meter.',
                 datatype='float'),
        XmlField(name='boormethode',
                 source_xpath='/methode',
                 definition='Boormethode voor het diepte-interval.',
                 datatype='string')
    ]


class Boring(AbstractDovType):
    """Class representing the DOV data type for boreholes."""

    subtypes = [BoorMethode]

    fields = [
        WfsField(name='pkey_boring', source_field='fiche', datatype='string'),
        WfsField(name='boornummer', source_field='boornummer',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        MvMtawField('Maaiveldhoogte in mTAW op dag dat de boring '
                    'uitgevoerd werd.'),
        WfsField(name='start_boring_mtaw', source_field='Z_mTAW',
                 datatype='float'),
        WfsField(name='gemeente', source_field='gemeente', datatype='string'),
        XmlField(name='diepte_boring_van',
                 source_xpath='/boring/diepte_van',
                 definition='Startdiepte van de boring (in meter).',
                 datatype='float',
                 notnull=True),
        WfsField(name='diepte_boring_tot', source_field='diepte_tot_m',
                 datatype='float'),
        WfsField(name='datum_aanvang', source_field='datum_aanvang',
                 datatype='date'),
        WfsField(name='uitvoerder', source_field='uitvoerder',
                 datatype='string'),
        XmlField(name='boorgatmeting',
                 source_xpath='/boring/boorgatmeting/uitgevoerd',
                 definition='Is er een boorgatmeting uitgevoerd (ja/nee).',
                 datatype='boolean')
    ]

    pkey_fieldname = 'fiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Boring (borehole), being a URI of the form
            `https://www.dov.vlaanderen.be/data/boring/<id>`.

        """
        super().__init__('boring', pkey)



# -*- coding: utf-8 -*-
"""Module containing the abstract search classes to retrieve DOV data."""

from itertools import chain
import datetime
import math
import warnings
import re

import owslib
import owslib.fes2
from owslib.etree import etree
from owslib.feature import get_schema
from owslib.fes2 import FilterRequest
from owslib.wfs import WebFeatureService
import pandas as pd
import numpy as np

import pydov
from pydov.types.fields import (_WfsInjectedField, GeometryReturnField,
                                ReturnFieldList)
from pydov.util import owsutil
from pydov.util.dovutil import build_dov_url, get_xsd_schema
from pydov.util.errors import (InvalidFieldError, InvalidSearchParameterError,
                               LayerNotFoundError, WfsGetFeatureError,
                               DataParseWarning)
from pydov.util.hooks import HookRunner
from pydov.util.net import LocalSessionThreadPool

# compile regex for matching datetime
re_datetime = re.compile(
    r'([0-9]{4}-[0-9]{2}-[0-9]{2}T'
    r'[0-9]{2}:[0-9]{2}:[0-9]{2})'
    r'(\.[0-9]+)?([\+\-][0-9]{2}:?[0-9]{2})?(Z?)')


class AbstractCommon(object):
    """Class grouping methods common to AbstractSearch and
    AbstractTypeCommon."""

    @classmethod
    def __strtobool(cls, val):
        """Convert a string representation of truth to true (1) or false (0).

        True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
        are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
        'val' is anything else.

        Parameters
        ----------
        val : str
            String representation to convert to boolean.

        Returns
        -------
        boolean
            The converted boolean value.

        Raises
        ------
        ValueError
            If the string cannot be converted to a boolean value.
        """
        val = val.lower()
        if val in ('y', 'yes', 't', 'true', 'on', '1'):
            return True
        elif val in ('n', 'no', 'f', 'false', 'off', '0'):
            return False
        else:
            raise ValueError(
                "Cannot convert truth value %r to boolean." % (val,))

    @classmethod
    def _typeconvert(cls, text, returntype):
        """Parse the text to the given returntype.

        Parameters
        ----------
        text : str
           Text to convert
        returntype : str
            Parse the text to this output datatype. One of
            `string`, `float`, `integer`, `date`, `datetime`, `boolean`.

        Returns
        -------
        str or float or int or bool or datetime.date or datetime.datetime
            Returns the parsed text converted to the type described by
            `returntype`.

        """
        if returntype == 'string':
            def typeconvert(x):
                return u'' + (x.strip())
        elif returntype == 'integer':
            def typeconvert(x):
                return int(x)
        elif returntype == 'float':
            def typeconvert(x):
                return float(x)
        elif returntype == 'date':
            def typeconvert(x):
                # Patch for Zulu-time issue of geoserver for WFS 1.1.0
                if x.endswith('Z'):
                    return datetime.datetime.strptime(x, '%Y-%m-%dZ').date() \
                        + datetime.timedelta(days=1)
                else:
                    return datetime.datetime.strptime(x, '%Y-%m-%d').date()
        elif returntype == 'datetime':
            def typeconvert(x):
                x_match = re_datetime.search(x)
                if x_match is None:
                    raise ValueError(f'Cannot parse datetime from value "{x}"')
                x_datetime, x_millisecs, x_tz, x_zulu = x_match.groups()

                fmt = '%Y-%m-%dT%H:%M:%S'
                val = x_datetime

                if x_millisecs is not None:
                    x_millisecs = int(x_millisecs[1:])
                    fmt += '.%f'
                    val += f'.{x_millisecs:0>6}'

                if x_tz is not None:
                    fmt += '%z'
                    val += x_tz

                dtime = datetime.datetime.strptime(val, fmt)
                if x_zulu == 'Z':
                    dtime += datetime.timedelta(hours=1)
                return dtime
        elif returntype == 'boolean':
            def typeconvert(x):
                return cls.__strtobool(x)
        elif returntype == 'geometry':
            def typeconvert(x):
                if isinstance(x, etree._Element):
                    if owsutil.has_geom_support():
                        import shapely.geometry
                        import pygml
                        return shapely.geometry.shape(
                            pygml.parse(etree.tostring(x[0]).decode('utf8')))
                    else:
                        # this shouldn't happen
                        return etree.tostring(x[0]).decode('utf8')
                return np.nan
        else:
            def typeconvert(x):
                return x

        try:
            return typeconvert(text)
        except ValueError as e:
            warnings.warn(
                f"Failed to convert data to correct datatype: {e}. Resulting "
                "dataframe will be incomplete.",
                DataParseWarning)
            return np.nan


class AbstractSearch(AbstractCommon):
    """Abstract search class grouping methods common to all DOV search
    classes. Not to be instantiated or used directly."""

    def __init__(self, layer, objecttype):
        """Initialisation.

        Parameters
        ----------
        layer : str
            WFS layer to use for searching and retrieving records.
        objecttype : pydov.types.AbstractDovType
            Subclass of AbstractDovType indicating the associated DOV datatype.

        """
        self._layer = layer
        self._type = objecttype

        self._fields = None
        self._wfs_fields = None
        self._geometry_column = None

        self._map_wfs_source_df = {}
        self._map_df_wfs_source = {}

        self._wfs = None
        self._wfs_schema = None
        self._wfs_namespace = None
        self._wfs_max_features = None
        self._md_metadata = None
        self._fc_featurecatalogue = None
        self._xsd_schemas = None

    def _get_wfs_endpoint(self):
        """Get the WFS endpoint URL to use for accessing the feature type.

        Returns
        -------
        str
            The WFS endpoint URL.
        """
        base_url = build_dov_url('geoserver/')
        workspace = self._layer.split(':')[0]
        return base_url + workspace + '/wfs'

    def _init_wfs(self):
        """Initialise the WFS service. If the WFS service is not
        instantiated yet, do so and save it in a static variable available
        to all subclasses and instances.
        """
        if self._wfs is None:
            wfs_endpoint_url = self._get_wfs_endpoint()

            capabilities = owsutil.get_wfs_capabilities(
                wfs_endpoint_url + '?request=GetCapabilities&version=2.0.0')

            self._wfs = WebFeatureService(
                url=wfs_endpoint_url, version="2.0.0", xml=capabilities)

            self._wfs_max_features = owsutil.get_wfs_max_features(capabilities)

    def _init_namespace(self):
        """Initialise the WFS namespace associated with the layer.

        Raises
        ------
        NotImplementedError
            This is an abstract method that should be implemented in a
            subclass.

        """
        if self._wfs_namespace is None:
            self._wfs_namespace = self._get_namespace()

    def _init_fields(self):
        """Initialise the fields and their metadata available in this search
        class.

        Raises
        ------
        NotImplementedError
            This is an abstract method that should be implemented in a
            subclass.

        """
        if self._fields is None:
            if self._wfs_schema is None:
                self._wfs_schema = self._get_schema()

            if self._md_metadata is None:
                self._md_metadata = self._get_remote_metadata()

            if self._md_metadata is not None and \
                    self._fc_featurecatalogue is None:
                csw_url = self._get_csw_base_url()
                fc_uuid = owsutil.get_featurecatalogue_uuid(self._md_metadata)
                if fc_uuid is not None:
                    self._fc_featurecatalogue = \
                        owsutil.get_remote_featurecatalogue(csw_url, fc_uuid)

            if self._xsd_schemas is None:
                self._xsd_schemas = self._get_remote_xsd_schemas()

            fields = self._build_fields(
                self._wfs_schema,
                self._fc_featurecatalogue,
                self._xsd_schemas)

            for field in fields.values():
                if field['name'] not in self._type.get_field_names(
                        include_wfs_injected=True):
                    self._type.fields.append(
                        _WfsInjectedField(name=field['name'],
                                          datatype=field['type']))

            self._fields = self._build_fields(
                self._wfs_schema,
                self._fc_featurecatalogue,
                self._xsd_schemas)

    def _get_layer(self):
        """Get the WFS metadata for the layer.

        Returns
        -------
        owslib.feature.wfs110.ContentMetadata
            Content metadata describing the layer.

        Raises
        ------
        pydov.util.errors.LayerNotFoundError
            When the layer of this search class could not be found in the
            associated WFS service.

        """
        self._init_wfs()

        if self._layer not in self._wfs.contents:
            raise LayerNotFoundError(
                'Layer {} could not be found'.format(self._layer))
        else:
            return self._wfs.contents[self._layer]

    def _get_schema(self):
        """Get the WFS schema (i.e. the output of the DescribeFeatureType
        request) of the layer.

        Returns
        -------
        schema : dict
            Schema associated with the layer.

        """
        self._init_wfs()
        layername = self._layer.split(':')[1] if ':' in self._layer else \
            self._layer
        return get_schema(
            build_dov_url('geoserver/wfs'), layername, '2.0.0')

    def _get_namespace(self):
        """Get the WFS namespace of the layer.

        Returns
        -------
        namespace : str
            The namespace associated with the WFS layer. This is the
            namespace of the fields of this layer, needed to parse the
            output of a GetFeature request.

        """
        self._init_wfs()
        return owsutil.get_namespace(self._wfs, self._layer)

    def _get_remote_metadata(self):
        """Request and parse the remote metadata associated with the layer.

        Returns
        -------
        owslib.iso.MD_Metadata or None
            Parsed remote metadata describing the WFS layer in more detail,
            in the ISO 19115/19139 format or None when no metadata could be
            found or parsed.

        """
        wfs_layer = self._get_layer()
        return owsutil.get_remote_metadata(wfs_layer)

    def _get_remote_xsd_schemas(self):
        """Request and parse the remote XSD schemas associated with this type.

        Returns
        -------
        list of etree.ElementTree
            List of parsed XSD schemas associated with this type.

        """
        xsd_schemas = [get_xsd_schema(i) for i in self._type.get_xsd_schemas()]
        return [etree.fromstring(i) for i in xsd_schemas if i is not None]

    def _get_csw_base_url(self):
        """Get the CSW base url for the remote metadata associated with the
        layer.

        Returns
        -------
        url : str
            Base URL of the CSW service where the remote metadata and feature
            catalogue can be requested.

        """
        wfs_layer = self._get_layer()
        return owsutil.get_csw_base_url(wfs_layer)

    @classmethod
    def _get_xsd_enum_values(cls, xsd_schemas, xml_field):
        """Get the distinct enum values from XSD schemas for a given XML field.

        Depending of the 'xsd_type' of the XML field, retrieve the distinct
        enum values and definitions from the XSD schemas.

        Parameters
        ----------
        xsd_schemas : list of etree.ElementTree
            List of parsed XSD schemas.
        xml_field : dict
            Dictionary describing the XML field, including a 'xsd_type' key
            linking the type to the enum type in (one of) the XSD schemas.

        Returns
        -------
        values : dict
            Dictionary containing the enum values as keys (in the datatype
            of the XML field) and the definitions as values.

        """
        values = None
        if xml_field.get('xsd_type', None):
            values = {}
            for schema in xsd_schemas:
                tree_values = schema.findall(
                    './/{{http://www.w3.org/2001/XMLSchema}}simpleType['
                    '@name="{}"]/'
                    '{{http://www.w3.org/2001/XMLSchema}}restriction/'
                    '{{http://www.w3.org/2001/XMLSchema}}enumeration'.format(
                        xml_field.get('xsd_type')))
                for e in tree_values:
                    value = cls._typeconvert(
                        e.get('value'), xml_field.get('type'))
                    values[value] = e.findtext(
                        './{http://www.w3.org/2001/XMLSchema}annotation/{'
                        'http://www.w3.org/2001/XMLSchema}documentation')
            if len(values) == 0:
                values = None
        return values

    def _build_fields(self, wfs_schema, feature_catalogue, xsd_schemas):
        """Build the dictionary containing the metadata about the available
        fields.

        Parameters
        ----------
        wfs_schema : dict
            The schema associated with the WFS layer.
        feature_catalogue : dict
            Dictionary with fields described in the feature catalogue,
            as returned by pydov.util.owsutil.get_remote_featurecatalogue.

        Returns
        -------
        fields : dict<str,dict>
            Dictionary containing the metadata of the available fields,
            where the metadata dictionary includes:

            name (str)
                The name of the field.

            definition (str)
                The definition of the field.

            type (str)
                The datatype of the field.

            notnull (boolean)
                Whether the field is mandatory (True) or can be null (False).

            cost (integer)
                The cost associated with the request of this field in the
                output dataframe.

        Raises
        ------
        RuntimeError
            When the defined fields of this type are invalid.

        """
        fields = {}
        self._wfs_fields = []
        self._geometry_column = wfs_schema.get('geometry_column', None)

        for f in self._type.get_fields(include_subtypes=False).values():
            if not isinstance(f, pydov.types.fields.AbstractField):
                raise RuntimeError(
                    "Type '{}' fields should be instances of "
                    "pydov.types.fields.AbstractField, found {}.".format(
                        self._type.__name__, str(type(f))))

        for f in self._type.get_fields(include_subtypes=True).values():
            if not isinstance(f, pydov.types.fields.AbstractField):
                raise RuntimeError(
                    "Fields of subtype of '{}' should be instances of "
                    "pydov.types.fields.AbstractField, found {}.".format(
                        self._type.__name__, str(type(f))))

        _map_wfs_datatypes = {
            'int': 'integer',
            'long': 'integer',
            'decimal': 'float',
            'double': 'float',
            'dateTime': 'datetime'
        }

        df_wfs_fields = self._type.get_fields(source=('wfs',)).values()
        for f in df_wfs_fields:
            self._map_wfs_source_df[f['sourcefield']] = f['name']
            self._map_df_wfs_source[f['name']] = f['sourcefield']

        for wfs_field in wfs_schema['properties'].keys():
            if feature_catalogue is not None and \
                    wfs_field in feature_catalogue['attributes']:
                fc_field = feature_catalogue['attributes'][wfs_field]
                self._wfs_fields.append(wfs_field)

                name = self._map_wfs_source_df.get(wfs_field, wfs_field)

                field = {
                    'name': name,
                    'definition': fc_field['definition'],
                    'type': _map_wfs_datatypes.get(
                        wfs_schema['properties'][wfs_field],
                        wfs_schema['properties'][wfs_field]),
                    'notnull': fc_field['multiplicity'][0] > 0,
                    'query': True,
                    'cost': 1
                }

                if fc_field['values'] is not None:
                    field['values'] = fc_field['values']

                fields[name] = field
            else:
                self._wfs_fields.append(wfs_field)

                name = self._map_wfs_source_df.get(wfs_field, wfs_field)

                field = {
                    'name': name,
                    'definition': None,
                    'type': _map_wfs_datatypes.get(
                        wfs_schema['properties'][wfs_field],
                        wfs_schema['properties'][wfs_field]),
                    'notnull': False,
                    'query': True,
                    'cost': 1
                }

                fields[name] = field

        if owsutil.has_geom_support() and 'geometry' in wfs_schema and \
                'geometry_column' in wfs_schema:
            name = wfs_schema['geometry_column']
            self._wfs_fields.append(name)

            field = {
                'name': name,
                'definition': None,
                'type': 'geometry',
                'notnull': False,
                'query': False,
                'cost': 1
            }

            fields[name] = field

        for xml_field in self._type.get_fields(source=['xml']).values():
            field = {
                'name': xml_field['name'],
                'type': xml_field['type'],
                'definition': xml_field['definition'],
                'notnull': xml_field['notnull'],
                'query': False,
                'cost': 10
            }

            values = self._get_xsd_enum_values(xsd_schemas, xml_field)
            if values is not None:
                field['values'] = values

            fields[field['name']] = field

        for custom_field in self._type.get_fields(
                source=['custom_wfs']).values():
            field = {
                'name': custom_field['name'],
                'type': custom_field['type'],
                'definition': custom_field['definition'],
                'notnull': custom_field['notnull'],
                'query': False,
                'cost': 1
            }
            fields[field['name']] = field

        for custom_field in self._type.get_fields(
                source=['custom_xml']).values():
            field = {
                'name': custom_field['name'],
                'type': custom_field['type'],
                'definition': custom_field['definition'],
                'notnull': custom_field['notnull'],
                'query': False,
                'cost': 10
            }
            fields[field['name']] = field

        return fields

    def _pre_search_validation(self, location, query, sort_by,
                               return_fields, max_features):
        """Perform validation on the parameters of the search query.

        Parameters
        ----------
        location : pydov.util.location.AbstractLocationFilter
            Location filter limiting the features to retrieve.
        query : owslib.fes2.OgcExpression
            OGC filter expression to use for searching. This can contain any
            combination of filter elements defined in owslib.fes2. The query
            should use the fields provided in `get_fields()`. Note that not
            all fields are currently supported as a search parameter.
        sort_by : owslib.fes2.SortBy, optional
            List of properties to sort by.
        return_fields : pydov.types.fields.ReturnFieldList
            A list of fields to be returned in the output data. This should
            be a subset of the fields provided in `get_fields()`.
        max_features : int
            Limit the maximum number of features to request.

        Raises
        ------
        pydov.util.errors.InvalidSearchParameterError
            When not one of `location`, `query` or `max_features` is provided.

        pydov.util.errors.InvalidFieldError
            When at least one of the fields in `return_fields` is unknown.

            When a field that is only accessible as return field is used as
            a query parameter.

        """
        if location is None and query is None and max_features is None:
            raise InvalidSearchParameterError(
                'Provide either the location or the query parameter or the '
                'max_features parameter.'
            )

        if query is not None:
            if not isinstance(query, owslib.fes2.OgcExpression):
                if isinstance(query, owslib.fes.OgcExpression):
                    raise InvalidSearchParameterError(
                        "Query should be an owslib.fes2.OgcExpression.\n"
                        "Try importing your query operators, like "
                        f"'{query.__class__.__name__}', "
                        "from the owslib.fes2 package."
                    )

                raise InvalidSearchParameterError(
                    "Query should be an owslib.fes2.OgcExpression.")

            filter_request = FilterRequest()
            filter_request = filter_request.setConstraint(query)

            self._init_fields()
            for property_name in filter_request.findall(
                    './/{http://www.opengis.net/fes/2.0}ValueReference'):
                name = property_name.text
                if name not in self._map_df_wfs_source \
                        and name not in self._wfs_fields:
                    if name in self._fields:
                        raise InvalidFieldError(
                            "Cannot use return field '{}' in query.".format(
                                name))
                    raise InvalidFieldError(
                        "Unknown query parameter: '{}'".format(name))

                if name in self._wfs_fields and name in self._fields and \
                        self._fields[name]['type'] == 'geometry':
                    raise InvalidFieldError(
                        ("Cannot use geometry field '{}' in attribute query. "
                         "Use the 'location' parameter for "
                         "spatial filtering.").format(name)
                    )

        if location is not None:
            self._init_fields()

        if sort_by is not None:
            if not isinstance(sort_by, owslib.fes2.SortBy):
                raise InvalidSearchParameterError(
                    "SortBy should be an owslib.fes2.SortBy")

            self._init_fields()
            for property_name in sort_by.toXML().findall(
                    './/{http://www.opengis.net/fes/2.0}ValueReference'):
                name = property_name.text
                if name not in self._map_df_wfs_source \
                        and name not in self._wfs_fields:
                    if name in self._fields:
                        raise InvalidFieldError(
                            "Cannot use return field '{}' for sorting.".format(
                                name))
                    raise InvalidFieldError(
                        "Unknown query parameter: '{}'".format(name))

        return_fields = ReturnFieldList.from_field_names(return_fields)
        if return_fields is not None:
            if not isinstance(return_fields, ReturnFieldList):
                raise AttributeError(
                    'return_fields should be an instance of '
                    'pydov.types.fields.ReturnFieldList')

            self._init_fields()
            for rf in return_fields:
                if rf.name not in self._fields:
                    if rf.name in self._map_wfs_source_df:
                        raise InvalidFieldError(
                            "Unknown return field: "
                            "'{}'. Did you mean '{}'?".format(
                                rf, self._map_wfs_source_df[rf]))
                    if rf.name.lower() in [i.lower() for i in
                                           self._map_wfs_source_df.keys()]:
                        sugg = [i for i in self._map_wfs_source_df.keys() if
                                i.lower() == rf.name.lower()][0]
                        raise InvalidFieldError(
                            "Unknown return field: "
                            "'{}'. Did you mean '{}'?".format(rf.name, sugg))
                    raise InvalidFieldError(
                        "Unknown return field: '{}'".format(rf.name))
        elif len(self._type.fields) == 0:
            self._init_fields()

    @staticmethod
    def _get_remote_wfs_feature(wfs, typename, location, filter,
                                sort_by, propertyname, max_features,
                                geometry_column, crs=None, start_index=0,
                                session=None):
        """Perform the WFS 2.0 GetFeature call to get features from the remote
        service.

        Parameters
        ----------
        typename : str
            Layername to query.
        location : pydov.util.location.AbstractLocationFilter
            Location filter limiting the features to retrieve.
        filter : str of owslib.fes2.FilterRequest
            Filter request to search on attribute values.
        sort_by : str of owslib.fes2.SortBy, optional
            List of properties to sort by.
        propertyname : list<str>
            List of properties to return.
        max_features : int
            Limit the maximum number of features to request.
        geometry_column : str
            Name of the geometry column to use in the spatial filter.
        crs : str
            EPSG code of the CRS of the geometries that will be returned.
            Defaults to None, which means the default CRS of the WFS layer.
        start_index : int
            Index of the first feature to return. Can be used for paging.
        session : requests.Session
            Session to use to perform HTTP requests for data. Defaults to None,
            which means a new session will be created for each request.

        Returns
        -------
        wfs_response, wfs_getfeature_request : bytes, etree.Element
            Response of the WFS service.

        """
        wfs_getfeature_xml = owsutil.wfs_build_getfeature_request(
            geometry_column=geometry_column,
            typename=typename,
            location=location,
            filter=filter,
            sort_by=sort_by,
            max_features=max_features,
            propertyname=propertyname,
            start_index=start_index,
            crs=crs
        )

        tree = HookRunner.execute_inject_wfs_getfeature_response(
            wfs_getfeature_xml)

        if tree is not None:
            return tree, wfs_getfeature_xml

        return owsutil.wfs_get_feature(
            baseurl=wfs.url,
            get_feature_request=wfs_getfeature_xml,
            session=session
        ), wfs_getfeature_xml

    def _search(self, location=None, query=None, return_fields=None,
                sort_by=None, max_features=None, extra_wfs_fields=[]):
        """Perform the WFS search by issuing a GetFeature request.

        Parameters
        ----------
        location : pydov.util.location.AbstractLocationFilter
            Location filter limiting the features to retrieve.
        query : owslib.fes2.OgcExpression
            OGC filter expression to use for searching. This can contain any
            combination of filter elements defined in owslib.fes2. The query
            should use the fields provided in `get_fields()`. Note that not
            all fields are currently supported as a search parameter.
        return_fields : list<str>
            A list of fields to be returned in the output data. This should
            be a subset of the fields provided in `get_fields()`. Note that
            not all fields are currently supported as return fields.
        sort_by : owslib.fes2.SortBy, optional
            List of properties to sort by.
        max_features : int
            Limit the maximum number of features to request.
        extra_wfs_fields: list<str>
            A list of extra fields to be included in the WFS requests,
            regardless whether they're needed as return field. Optional,
            defaults to an empty list.

        Returns
        ------
        list of etree.Element
            XML trees of the WFS responses containing the features matching
            the location and the query.

        Raises
        ------
        pydov.util.errors.InvalidSearchParameterError
            When not one of `location`, `query` or `max_features` is provided.

        pydov.util.errors.InvalidFieldError
            When at least one of the fields in `return_fields` is unknown.

            When a field that is only accessible as return field is used as
            a query parameter.

            When a field that can only be used as a query parameter is used as
            a return field.

        """
        self._pre_search_validation(location, query, sort_by, return_fields,
                                    max_features)
        self._init_namespace()
        self._init_wfs()

        filter_request = None
        if query is not None:
            filter_request = FilterRequest()
            filter_request = filter_request.setConstraint(query)

        if filter_request is not None:
            for property_name in filter_request.findall(
                    './/{http://www.opengis.net/fes/2.0}ValueReference'):
                property_name.text = self._map_df_wfs_source.get(
                    property_name.text, property_name.text)

            filter_request = etree.tostring(filter_request, encoding='unicode')

        if self._type.pkey_fieldname is not None:
            wfs_property_names = [self._type.pkey_fieldname]
        else:
            wfs_property_names = []

        if return_fields is None:
            wfs_property_names.extend([
                f['sourcefield'] for f in self._type.get_fields(
                    source=('wfs',)).values() if (
                        self._type.pkey_fieldname is None
                        or not f.get('wfs_injected', False))])
            geom_return_crs = None
        else:
            wfs_property_names.extend([self._map_df_wfs_source[i]
                                       for i in self._map_df_wfs_source
                                       if i in return_fields])

            geom_return_crs = [f.epsg for f in return_fields
                               if isinstance(f, GeometryReturnField)]
            if len(geom_return_crs) > 0:
                geom_return_crs = f'EPSG:{geom_return_crs[0]}'
            else:
                geom_return_crs = None

        extra_custom_fields = set()
        for custom_field in self._type.get_fields(
                source=('custom_wfs',)).values():
            extra_custom_fields.update(custom_field.requires_wfs_fields())

        wfs_property_names.extend(extra_wfs_fields)
        wfs_property_names.extend(list(extra_custom_fields))
        wfs_property_names = list(set(wfs_property_names))

        if sort_by is not None:
            sort_by_xml = sort_by.toXML()
            for property_name in sort_by_xml.findall(
                    './/{http://www.opengis.net/fes/2.0}ValueReference'):
                property_name.text = self._map_df_wfs_source.get(
                    property_name.text, property_name.text)

            sort_by = etree.tostring(sort_by_xml, encoding='unicode')

        HookRunner.execute_wfs_search_init(params=dict(
            typename=self._layer,
            location=location,
            filter=filter_request,
            sort_by=sort_by,
            max_features=max_features,
            propertynames=wfs_property_names,
            geometry_column=self._geometry_column
        ))

        def _get_remote_wfs(start_index=0, max_features=None, session=None):
            fts, getfeature = self._get_remote_wfs_feature(
                wfs=self._wfs,
                typename=self._layer,
                location=location,
                filter=filter_request,
                sort_by=sort_by,
                max_features=max_features,
                propertyname=wfs_property_names,
                geometry_column=self._geometry_column,
                crs=geom_return_crs,
                start_index=start_index,
                session=session)

            tree = etree.fromstring(fts)

            if tree.get('numberReturned') is None:
                raise WfsGetFeatureError(
                    "Error retrieving features of layer '{}' from "
                    "DOV WFS server:\n{}".format(
                        self._layer,
                        etree.tostring(tree).decode('utf8')))

            number_matched = int(tree.get('numberMatched'))
            number_returned = int(tree.get('numberReturned'))

            HookRunner.execute_wfs_search_result(
                number_matched, number_returned)

            HookRunner.execute_wfs_search_result_received(getfeature, tree)

            return tree

        result = []

        # execute the first WFS query
        tree = _get_remote_wfs(
            start_index=0, max_features=max_features, session=pydov.session)
        result.append(tree)

        number_matched = int(tree.get('numberMatched'))
        number_returned = int(tree.get('numberReturned'))

        if max_features is not None and number_returned == max_features:
            # we asked for a limited number of features and we got all of them,
            # we're done!
            pass
        elif number_matched == number_returned:
            # all features that matched the query were returned,
            # we're done!
            pass
        else:
            # more features matched the query than were returned by the server,
            # we need more requests to fetch the rest of the results
            pool = LocalSessionThreadPool()

            if max_features is not None:
                fts_to_get = min(
                    max_features, number_matched) - number_returned
            else:
                fts_to_get = number_matched - number_returned
            fts_per_req = self._wfs_max_features or number_returned
            extra_reqs = math.ceil(fts_to_get/fts_per_req)

            for i in range(extra_reqs):
                start_index = (i+1)*fts_per_req
                if i == extra_reqs - 1:
                    # last request
                    if fts_to_get == fts_per_req:
                        max_features = None
                    else:
                        max_features = fts_to_get % fts_per_req
                else:
                    max_features = fts_per_req

                pool.execute(_get_remote_wfs, (start_index, max_features))

            for r in pool.join():
                if r.get_error():
                    raise r.get_error()

                worker_result = r.get_result()
                if worker_result is not None and len(worker_result) > 0:
                    result.append(worker_result)

        return result

    def get_description(self):
        """Get the description of this search layer.

        Returns
        -------
        str
            The description of this layer.

        """
        wfs_layer = self._get_layer()
        return wfs_layer.abstract

    def get_fields(self):
        """Get the metadata of the fields that are available.

        Returns
        -------
        fields : dict<str,dict>
            Dictionary containing the metadata of the available fields,
            where the metadata dictionary includes:

            name (str)
                The name of the field.

            definition (str)
                The definition of the field.

            type (str)
                The datatype of the field.

            notnull (boolean)
                Whether the field is mandatory (True) or can be null (False).

            query (boolean)
                Whether the field can be used in an attribute query.

            cost (integer)
                The cost associated with the request of this field in the
                output dataframe.

            Optionally, it can contain:

            values (list)
                A list of possible values for this field.

        """
        self._init_fields()
        return self._fields

    def search(self, location=None, query=None,
               sort_by=None, return_fields=None, max_features=None):
        """Search for objects of this type. Provide `location` and/or
        `query` and/or `max_features`.
        When `return_fields` is None, all fields are returned.

        Parameters
        ----------
        location : pydov.util.location.AbstractLocationFilter or \
                   owslib.fes2.BinaryLogicOpType<AbstractLocationFilter> or \
                   owslib.fes2.UnaryLogicOpType<AbstractLocationFilter>
            Location filter limiting the features to retrieve. Can either be a
            single instance of a subclass of AbstractLocationFilter, or a
            combination using And, Or, Not of AbstractLocationFilters.
        query : owslib.fes2.OgcExpression
            OGC filter expression to use for searching. This can contain any
            combination of filter elements defined in owslib.fes2. The query
            should use the fields provided in `get_fields()`. Note that not
            all fields are currently supported as a search parameter.
        sort_by : owslib.fes2.SortBy, optional
            List of properties to sort by.
        return_fields : list<str> or tuple<str> or set<str>
            A list of fields to be returned in the output data. This should
            be a subset of the fields provided in `get_fields()`. Note that
            not all fields are currently supported as return fields.
        max_features : int
            Limit the maximum number of features to request.

        Returns
        -------
        pandas.core.frame.DataFrame
            DataFrame containing the output of the search query.

        Raises
        ------
        pydov.util.errors.InvalidSearchParameterError
            When not one of `location` or `query` or `max_features` is
            provided.

        pydov.util.errors.InvalidFieldError
            When at least one of the fields in `return_fields` is unknown.

            When a field that is only accessible as return field is used as
            a query parameter.

            When a field that can only be used as a query parameter is used as
            a return field.

        AttributeError
            When the argument supplied as return_fields is not a list,
            tuple or set.

        NotImplementedError
            This is an abstract method that should be implemented in a
            subclass.

        """
        return_fields = ReturnFieldList.from_field_names(return_fields)

        trees = self._search(location=location, query=query, sort_by=sort_by,
                             return_fields=return_fields,
                             max_features=max_features)

        feature_generators = []
        for tree in trees:
            feature_generators.append(
                self._type.from_wfs(tree, self._wfs_namespace))

        cols = self._type.get_field_names(return_fields, include_geometry=True)
        if len(cols) == 0:
            cols = self._type.get_field_names(
                return_fields, include_wfs_injected=True,
                include_geometry=False)

        df = pd.DataFrame(
            data=self._type.to_df_array(
                chain.from_iterable(feature_generators), return_fields),
            columns=cols)
        return df



# -*- coding: utf-8 -*-
"""Module grouping tests for the pydov.types.bodemobservatie module."""

from pydov.types.bodemobservatie import Bodemobservatie
from pydov.types.fields import ReturnFieldList
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/bodemobservatie/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemobservatie/feature.xml'
location_dov_xml = 'tests/data/types/bodemobservatie/bodemobservatie.xml'


class TestBodemobservatie(AbstractTestTypes):
    """Class grouping tests for the
    pydov.types.bodemobservatie.Bodemobservatie class."""

    datatype_class = Bodemobservatie
    namespace = 'https://www.dov.vlaanderen.be/bodem'
    pkey_base = build_dov_url('data/bodemobservatie/')

    field_names = ['pkey_bodemobservatie', 'pkey_bodemlocatie',
                   'pkey_parent', 'x', 'y', 'mv_mtaw', 'diepte_van_cm',
                   'diepte_tot_cm', 'observatiedatum', 'invoerdatum',
                   'parametergroep', 'parameter', 'detectie', 'waarde',
                   'eenheid', 'veld_labo', 'methode', 'betrouwbaarheid',
                   'fractiemeting_ondergrens', 'fractiemeting_bovengrens',
                   'fractiemeting_waarde']
    field_names_subtypes = []
    field_names_nosubtypes = [
        'pkey_bodemobservatie', 'pkey_bodemlocatie',
        'pkey_parent', 'x', 'y', 'mv_mtaw', 'diepte_van_cm',
        'diepte_tot_cm', 'observatiedatum', 'invoerdatum',
        'parametergroep', 'parameter', 'detectie', 'waarde',
        'eenheid', 'veld_labo', 'methode', 'betrouwbaarheid']

    valid_returnfields = ReturnFieldList.from_field_names('pkey_bodemobservatie', 'parameter')
    valid_returnfields_subtype = ReturnFieldList.from_field_names('pkey_bodemobservatie', 'parameter',
                                                                  'fractiemeting_ondergrens')

    inexistent_field = 'onbestaand'



"""Tests for grondwatervergunningen"""
from owslib.fes2 import PropertyIsEqualTo
from pandas import DataFrame

from pydov.search.grondwatervergunning import GrondwaterVergunningSearch
from pydov.types.fields import ReturnFieldList
from pydov.types.grondwatervergunning import GrondwaterVergunning
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = \
    'tests/data/types/grondwatervergunning/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondwatervergunning/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondwatervergunning/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/grondwatervergunning/' \
                          'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/grondwatervergunning/feature.xml'
location_dov_xml = None
location_xsd_base = 'tests/data/types/grondwatervergunning/xsd_*.xml'


class TestGrondwaterVergunningSearch(AbstractTestSearch):

    search_instance = GrondwaterVergunningSearch()
    datatype_class = GrondwaterVergunning

    valid_query_single = PropertyIsEqualTo(
        propertyname='installatie',
        literal=build_dov_url('data/installatie/2019-088045'))

    inexistent_field = 'onbestaand'
    xml_field = None
    wfs_field = 'installatie'

    valid_returnfields = ReturnFieldList.from_field_names('id_vergunning', 'diepte')
    valid_returnfields_subtype = None
    valid_returnfields_extra = ReturnFieldList.from_field_names('inrichtingsnummer', 'vergund_aantal_putten')

    df_default_columns = [
        'id_vergunning', 'pkey_installatie', 'x', 'y',
        'diepte', 'exploitant_naam', 'watnr', 'vlaremrubriek',
        'vergund_jaardebiet', 'vergund_dagdebiet',
        'van_datum_termijn', 'tot_datum_termijn',
        'aquifer_vergunning', 'inrichtingsklasse', 'nacebelcode',
        'actie_waakgebied', 'cbbnr', 'kbonr']

    def test_search_customreturnfields(self, mp_get_schema,
                                       mp_remote_describefeaturetype,
                                       mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with custom return fields.

        Test whether the output dataframe is correct.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType .
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('id_vergunning', 'pkey_installatie'))

        assert isinstance(df, DataFrame)

        assert list(df) == ['id_vergunning', 'pkey_installatie']

    def test_search_wfs_resolve(self, mp_get_schema,
                                mp_remote_describefeaturetype,
                                mp_remote_wfs_feature,):
        """Test the search method with return fields from WFS but not from
        XML.

        Test whether the output dataframe contains the resolved WFS data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('id_vergunning', 'diepte'))

        assert df.diepte[0] == 32.0



"""Module grouping online tests for the generic search module."""

import pytest

from pydov.search.generic import WfsSearch
from tests.abstract import ServiceCheck


class TestWfsSearchOnline:
    """Class grouping online tests for the generic search module. """

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_multiple_types(self):
        """Test whether multiple instances of WfsSearch don't share their
        fields.
        """
        search1 = WfsSearch('dov-pub:Opdrachten')
        search2 = WfsSearch('erosie:erosie_gemeente')

        assert search1.get_fields() != search2.get_fields()



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve permit data."""
from pydov.search.abstract import AbstractSearch
from pydov.types.grondwatervergunning import GrondwaterVergunning


class GrondwaterVergunningSearch(AbstractSearch):
    """Search class to retrieve information about Gw permits (
    all permits: current and historical)."""

    def __init__(self, objecttype=GrondwaterVergunning):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GrondwaterVergunning type.
            Optional: defaults to the GrondwaterVergunning type containing
            the fields described in the documentation.

        """
        super(GrondwaterVergunningSearch, self).__init__(
            'gw_vergunningen:alle_verg', objecttype)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV bodemsite data."""
from ..types.bodemsite import Bodemsite
from .abstract import AbstractSearch


class BodemsiteSearch(AbstractSearch):
    """Search class to retrieve information about bodemlocaties."""

    def __init__(self, objecttype=Bodemsite):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemsite type.
            Optional: defaults to the Bodemsite type containing the fields
            described in the documentation.

        """
        super(BodemsiteSearch, self).__init__('bodem:bodemsites', objecttype)



# -*- coding: utf-8 -*-
"""Module containing generic search classes to retrieve DOV WFS data."""

from pydov.search.abstract import AbstractSearch
from pydov.types.generic import WfsTypeFactory


class WfsSearch(AbstractSearch):
    """Search class for generic WFS layers from DOV."""

    def __init__(self, layer):
        """Initialisation.

        Create a new WfsSearch instance to query a generic WFS layer from DOV.
        Check out the available WFS layers via
        https://dov.vlaanderen.be/geonetwork

        Parameters
        ----------
        layer : str
            Workspace qualified layername of the layer to query.
        """
        super(WfsSearch, self).__init__(
            layer, WfsTypeFactory.get_wfs_type(layer))



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV groundwater screen
 data."""
from owslib.fes2 import And, Not, PropertyIsNull

from ..types.grondwaterfilter import GrondwaterFilter
from .abstract import AbstractSearch


class GrondwaterFilterSearch(AbstractSearch):
    """Search class to retrieve information about groundwater screens
    (GrondwaterFilter).
    """

    def __init__(self, objecttype=GrondwaterFilter):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GrondwaterFilter type.
            Optional: defaults to the GrondwaterFilter type containing the
            fields described in the documentation.

        """
        super(GrondwaterFilterSearch,
              self).__init__('gw_meetnetten:meetnetten', objecttype)

    def search(self, location=None, query=None, sort_by=None,
               return_fields=None, max_features=None):
        """Search for objects of this type. Provide `location` and/or
        `query` and/or `max_features`.
        When `return_fields` is None, all fields are returned.

        Excludes 'empty' filters (i.e. Putten without Filters) by extending
        the `query` with a not-null check on pkey_filter.

        Parameters
        ----------
        location : pydov.util.location.AbstractLocationFilter or \
                   owslib.fes2.BinaryLogicOpType<AbstractLocationFilter> or \
                   owslib.fes2.UnaryLogicOpType<AbstractLocationFilter>
            Location filter limiting the features to retrieve. Can either be a
            single instance of a subclass of AbstractLocationFilter, or a
            combination using And, Or, Not of AbstractLocationFilters.
        query : owslib.fes2.OgcExpression
            OGC filter expression to use for searching. This can contain any
            combination of filter elements defined in owslib.fes2. The query
            should use the fields provided in `get_fields()`. Note that not
            all fields are currently supported as a search parameter.
        sort_by : owslib.fes2.SortBy, optional
            List of properties to sort by.
        return_fields : list<str> or tuple<str> or set<str>
            A list of fields to be returned in the output data. This should
            be a subset of the fields provided in `get_fields()`. Note that
            not all fields are currently supported as return fields.
        max_features : int
            Limit the maximum number of features to request.

        Returns
        -------
        pandas.core.frame.DataFrame
            DataFrame containing the output of the search query.

        Raises
        ------
        pydov.util.errors.InvalidSearchParameterError
            When not one of `location` or `query` or `max_features` is
            provided.

        pydov.util.errors.InvalidFieldError
            When at least one of the fields in `return_fields` is unknown.

            When a field that is only accessible as return field is used as
            a query parameter.

            When a field that can only be used as a query parameter is used as
            a return field.

        AttributeError
            When the argument supplied as return_fields is not a list,
            tuple or set.

        NotImplementedError
            This is an abstract method that should be implemented in a
            subclass.

        """
        self._pre_search_validation(location, query, sort_by, return_fields,
                                    max_features)

        exclude_empty_filters = Not([PropertyIsNull(
                                     propertyname='pkey_filter')])

        if query is not None:
            query = And([query, exclude_empty_filters])
        else:
            query = exclude_empty_filters

        return super().search(
            location=location, query=query, sort_by=sort_by,
            return_fields=return_fields, max_features=max_features)



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemobservaties, including
subtypes."""
from pydov.types.fields import WfsField, XmlField

from .abstract import AbstractDovSubType, AbstractDovType


class Fractiemeting(AbstractDovSubType):
    rootpath = './/bodemobservatie/fractiemeting'

    fields = [
        XmlField(name='fractiemeting_ondergrens',
                 source_xpath='ondergrens',
                 definition='ondergrens van de fractiemeting in µm',
                 datatype='float',
                 notnull=False),
        XmlField(name='fractiemeting_bovengrens',
                 source_xpath='bovengrens',
                 definition='bovengrens van de fractiemeting in µm',
                 datatype='float',
                 notnull=False),
        XmlField(name='fractiemeting_waarde',
                 source_xpath='waarde',
                 definition='gemeten massaprocent van de fractie',
                 datatype='float',
                 notnull=False)
    ]


class Bodemobservatie(AbstractDovType):
    """Class representing the DOV data type for bodemobservaties."""

    subtypes = [Fractiemeting]

    fields = [
        WfsField(name='pkey_bodemobservatie',
                 source_field='Bodemobservatiefiche',
                 datatype='string'),
        WfsField(name='pkey_bodemlocatie',
                 source_field='Bodemlocatiefiche',
                 datatype='string'),
        WfsField(name='pkey_parent',
                 source_field='Parentfiche',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='mv_mtaw', source_field='mv_mTAW', datatype='float'),
        WfsField(name='diepte_van_cm', source_field='Diepte_van',
                 datatype='float'),
        WfsField(name='diepte_tot_cm', source_field='Diepte_tot',
                 datatype='float'),
        XmlField(name='observatiedatum',
                 source_xpath='/bodemobservatie/observatiedatum',
                 definition='Observatiedatum van de bodemobservatie',
                 datatype='date'),
        XmlField(name='invoerdatum',
                 source_xpath='/bodemobservatie/invoerdatum',
                 definition='Invoerdatum van de bodemobservatie.',
                 datatype='date'),
        XmlField(name='parametergroep',
                 source_xpath='/bodemobservatie/parametergroep',
                 definition='Indeling van de parameter naar groep, '
                 + 'bvb anionen, kationen, .... Is indicatief.',
                 datatype='string'),
        WfsField(name='parameter', source_field='Parameter',
                 datatype='string'),
        XmlField(name='detectie',
                 source_xpath='/bodemobservatie/detectieconditie',
                 definition='boven/onder detectielimiet',
                 datatype='string'),
        WfsField(name='waarde', source_field='Waarde', datatype='string'),
        WfsField(name='eenheid', source_field='Eenheid', datatype='string'),
        WfsField(name='veld_labo', source_field='Labo_of_veld',
                 datatype='string'),
        WfsField(name='methode', source_field='Observatiemethode',
                 datatype='string'),
        XmlField(name='betrouwbaarheid',
                 source_xpath='/bodemobservatie/betrouwbaarheid',
                 definition='Betrouwbaarheid van de meting.',
                 datatype='string')
    ]

    pkey_fieldname = 'Bodemobservatiefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemobservatie, being a URI of the form
            `https://www.dov.vlaanderen.be/data/bodemobservatie/<id>`.

        """
        super().__init__('bodemobservatie', pkey)
